/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.framework;

import java.io.File;
import java.io.InputStream;
import java.util.List;

import android.app.Activity;
import android.app.Application;
import android.app.Dialog;
import android.content.Intent;
import android.content.pm.ApplicationInfo;
import android.content.res.Resources;
import android.net.Uri;
import android.os.Build;
import android.os.Looper;
import android.taobao.atlas.runtime.ActivityTaskMgr;
import android.taobao.atlas.runtime.SecurityHandler;
import android.taobao.atlas.util.ApkUtils;
import android.taobao.atlas.util.WrapperUtil;
import android.text.TextUtils;
import android.util.Log;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;
import org.osgi.framework.BundleListener;

import android.taobao.atlas.hack.AndroidHack;
import android.taobao.atlas.hack.AssertionArrayException;
import android.taobao.atlas.hack.AtlasHacks;
import android.taobao.atlas.runtime.ActivityManagerDelegate;
import android.taobao.atlas.runtime.BundleLifecycleHandler;
import android.taobao.atlas.runtime.ClassNotFoundInterceptorCallback;
import android.taobao.atlas.runtime.DelegateClassLoader;
import android.taobao.atlas.runtime.FrameworkLifecycleHandler;
import android.taobao.atlas.runtime.InstrumentationHook;
import android.taobao.atlas.runtime.RuntimeVariables;

import static android.taobao.atlas.runtime.InstrumentationHook.sOnIntentRedirectListener;

public class Atlas {

    public static String sAPKSource ;
    protected static Atlas instance;
    public static boolean Downgrade_H5 = false;
    public static boolean isDebug;

    private Atlas(){
    }

    public static synchronized Atlas getInstance() {
        if (instance == null) {
            instance = new Atlas();
        }
        return instance;
    }

    private BundleLifecycleHandler    bundleLifecycleHandler;
    private FrameworkLifecycleHandler frameworkLifecycleHandler;

    /**
     * Init the framework
     * 
     * @param application
     * @return
     * @throws Exception
     */
    public void init(Application application,boolean reset) throws AssertionArrayException, Exception {
        if(application==null){
            throw new RuntimeException("application is null");
        }
        ApplicationInfo app_info = application.getApplicationInfo();
        sAPKSource = app_info.sourceDir;
        boolean DEBUG = (app_info.flags & ApplicationInfo.FLAG_DEBUGGABLE) != 0;
        RuntimeVariables.androidApplication = application;
        RuntimeVariables.delegateResources  = application.getResources();
        Framework.containerVersion = RuntimeVariables.sInstalledVersionName;
        ClassLoader cl = Atlas.class.getClassLoader();
        Framework.systemClassLoader = cl;
        // defineAndVerify
        String packageName = application.getPackageName();
        // 
        DelegateClassLoader newClassLoader = new DelegateClassLoader(cl);
        // init RuntimeVariables
        RuntimeVariables.delegateClassLoader = newClassLoader;

        AndroidHack.injectClassLoader(packageName, newClassLoader);
        AndroidHack.injectInstrumentationHook(new InstrumentationHook(AndroidHack.getInstrumentation(), application.getBaseContext()));
        // add listeners
        bundleLifecycleHandler = new BundleLifecycleHandler();
        Framework.syncBundleListeners.add(bundleLifecycleHandler);
        frameworkLifecycleHandler = new FrameworkLifecycleHandler();
        Framework.frameworkListeners.add(frameworkLifecycleHandler);

        try {
            ActivityManagerDelegate activityManagerProxy = new ActivityManagerDelegate();

            Object gDefault = null;
            if(Build.VERSION.SDK_INT>25 || (Build.VERSION.SDK_INT==25&&Build.VERSION.PREVIEW_SDK_INT>0)){
                gDefault=AtlasHacks.ActivityManager_IActivityManagerSingleton.get(AtlasHacks.ActivityManager.getmClass());
            }else{
                gDefault=AtlasHacks.ActivityManagerNative_gDefault.get(AtlasHacks.ActivityManagerNative.getmClass());
            }
            AtlasHacks.Singleton_mInstance.hijack(gDefault, activityManagerProxy);
        }catch(Throwable e){}
        AndroidHack.hackH();
    }

    public void startup(Application application,boolean isUpdated) {
        if(!RuntimeVariables.safeMode) {
            if(application.getPackageManager().equals("com.taobao.taobao")) {
                try {
                    RuntimeVariables.sDexLoadBooster.getClass().getDeclaredMethod("setVerificationEnabled").invoke(RuntimeVariables.sDexLoadBooster, false);
                } catch (Throwable e) {
                    e.printStackTrace();
                }
            }
            if (!WrapperUtil.isDebugMode(application) && ApkUtils.isRootSystem()) {
                Atlas.getInstance().addBundleListener(new SecurityHandler());
            }
            try {
                Framework.startup(isUpdated);
            } catch (Exception e) {
                throw new RuntimeException( e);
            }
            if(RuntimeVariables.sCurrentProcessName.equals(RuntimeVariables.androidApplication.getPackageName())) {
                System.setProperty("BUNDLES_INSTALLED", "true");
                application.getBaseContext().sendBroadcast(new Intent("com.taobao.taobao.action.BUNDLES_INSTALLED"));
            }
        }
    }

    public void checkDownGradeToH5(Intent intent) {
        if (Downgrade_H5) {
            if (intent != null && intent.getComponent() != null) {
                intent.setComponent(null);
            }
            String url = intent.getDataString();
            if(!TextUtils.isEmpty(url)){
                if(url.contains("?")){
                    url = url+"&hybrid=true";
                }else{
                    url = url+"?hybrid=true";
                }
            }
            intent.setData(Uri.parse(url));
            intent.addCategory("com.taobao.intent.category.HYBRID_UI");
        }
    }

    private void checkingThread(boolean strict){
        if(Thread.currentThread().getId()==Looper.getMainLooper().getThread().getId() && Framework.isDeubgMode()){
            if(strict) {
                throw new RuntimeException("can not install bundle in ui thread");
            }else{
                Log.w("Atlas","can not install bundle in ui thread");
            }
        }
    }

    public void setIntentRedirectListener(InstrumentationHook.OnIntentRedirectListener listener){
        sOnIntentRedirectListener = listener;
    }

     ///////////////////////////////////////////////Deprecated method////////////////////////////////////////////
    /**
     *
     * install bundle from internal
     * @param location
     */
    @Deprecated
    public void installBundleWithDependency(String location){
        if (TextUtils.isEmpty(location)){
            Log.e("Atlas","empty location");
            return;
        }
        if (Framework.getBundle(location) == null){
            checkingThread(false);
            BundleInstallerFetcher.obtainInstaller().installTransitivelySync(new String[]{location});
        }
    }

    @Deprecated
    public void installBundle(final String location, final InputStream input) throws BundleException {
        if (TextUtils.isEmpty(location)){
            Log.e("Atlas","empty location");
        }
        if (Framework.getBundle(location) == null) {
            checkingThread(false);
            BundleInstallerFetcher.obtainInstaller().installSync(new String[]{location}, new InputStream[]{input});
        }
    }

    @Deprecated
    public void installBundle(final String location, final File file) throws BundleException {
        if (TextUtils.isEmpty(location)){
            Log.e("Atlas","empty location");
        }
        if (Framework.getBundle(location) == null) {
            checkingThread(false);
            BundleInstallerFetcher.obtainInstaller().installSync(new String[]{location}, new File[]{file});
        }
    }

    @Deprecated
    public void uninstallBundle(final String location) throws BundleException {
        Bundle bundle = Framework.getBundle(location);
        if (bundle != null) {
            BundleImpl b = (BundleImpl)bundle;
            File delDir = null;
            try {
                File soFile = b.getArchive().getArchiveFile();
                if(soFile.canWrite()){
                    soFile.delete();
                }
                delDir = b.getArchive().getCurrentRevision().getRevisionDir();
                bundle.uninstall();
                if(delDir !=null ){
                    Framework.deleteDirectory(delDir);
                }
            } catch (Exception e) {
            }

        } else {
            throw new BundleException("Could not uninstall bundle " + location + ", because could not find it");
        }
    }

    @Deprecated
    public Resources getDelegateResources() {
        return RuntimeVariables.delegateResources;
    }
    @Deprecated
    public ClassLoader getDelegateClassLoader() {
        return RuntimeVariables.delegateClassLoader;
    }
    ///////////////////////////////////////////////Deprecated method////////////////////////////////////////////


    /*************************************************↓↓↓↓↓↓for public use↓↓↓↓↓↓*************************************************************
     **************************************************↓↓↓↓↓↓for public use↓↓↓↓↓↓*************************************************************
     **************************************************↓↓↓↓↓↓for public use↓↓↓↓↓↓***************************************************************/

    /**
     * 同步安装一批apk内部的bundle
     * @param locations bundle的packagename
     * @param listener
     */
    public void installBundleTransitivelyAsync(String[] locations,BundleInstaller.InstallListener listener){
        BundleInstallerFetcher.obtainInstaller().installTransitivelyAsync(locations,listener);
    }

    /**
     * idle时候安装一批低优先级的apk内部的bundle
     * @param location bundle的packagename
     * @param listener
     */
    public void installIdleBundleTransitively(String location,BundleInstaller.InstallListener listener){
        BundleInstaller.startIdleInstall(location,listener);
    }

    /**
     * 相对空闲时候安装一批低优先级的apk内部的bundle
     * @param location bundle的packagename
     * @param listener
     */
    public void installDelayBundleTransitively(String location,BundleInstaller.InstallListener listener){
        BundleInstaller.startDelayInstall(location,listener);
    }

    /**
     *
     * @param location get installed bundle from Framework
     * @return
     */
    public Bundle getBundle(final String location) {
        return Framework.getBundle(location);
    }

    /**
     * 获取当前安装的所有bundle
     * @return
     */
    public List<Bundle> getBundles() {
        return Framework.getBundles();
    }

    /**
     * 获取bundle对应的classloader
     * @param location
     * @return
     */
    public ClassLoader getBundleClassLoader(final String location) {
        Bundle bundle = Framework.getBundle(location);
        if (bundle != null) {
            return ((BundleImpl) bundle).getClassLoader();
        }
        return null;
    }

    /**
     * 获取bundle对应的文件
     * @param location
     * @return
     */
    public File getBundleFile(final String location) {
        Bundle bundle = Framework.getBundle(location);
        if (bundle != null) {
            return ((BundleImpl) bundle).archive.getArchiveFile();
        }
        return null;
    }

    /**
     * 注册bundle生命周期监听
     * @param listener
     */
    public void addBundleListener(BundleListener listener) {
    	Framework.addBundleListener(listener);
    }

    /**
     * 注销bundle生命周期监听
     * @param listener
     */
    public void removeBundleListener(BundleListener listener) {
    	Framework.removeBundleListener(listener);
    }

    /**
     * Activity类查找失败监听(用于远程bundle和容错)
     * @param callback
     */
    public void setClassNotFoundInterceptorCallback(ClassNotFoundInterceptorCallback callback){
    	Framework.setClassNotFoundCallback(callback);
    }

    /**
     * 设置bundle运行时依赖
     * @param source  需要添加依赖的bundle
     * @param dependency 被依赖bundle的classloader
     * @param resourceDependencyNeed 是否需要使用被依赖bundle的资源
     */
    public void requestRuntimeDependency(ClassLoader source, ClassLoader dependency,boolean resourceDependencyNeed){
        if(source == getClass().getClassLoader()){
            throw new IllegalArgumentException("PathClassLoader can not have bundle dependency");
        }
        if(dependency == getClass().getClassLoader()){
            //bundle can use main dex class by default
            return;
        }
        if(!(source instanceof BundleClassLoader)){
            throw new IllegalArgumentException("source must be bundleclassloader");
        }
        if(!(dependency instanceof BundleClassLoader)){
            throw new IllegalArgumentException("dependency must be bundleclassloader");
        }
        String dependencyLocation = ((BundleClassLoader)dependency).location;
        ((BundleClassLoader)source).addRuntimeDependency(dependencyLocation);

        if(resourceDependencyNeed) {
            ActivityTaskMgr.getInstance().updateBundleActivityResource(dependencyLocation);
        }
    }

    /**
     * 设置bundle运行时依赖
     * @param source  需要添加依赖的bundle
     * @param dependencyBundle 被依赖的bundle的packagename
     * @param resourceDependencyNeed 是否需要使用被依赖bundle的资源
     * @throws BundleException
     */
    public void requestRuntimeDependency(ClassLoader source,String dependencyBundle,boolean resourceDependencyNeed) throws BundleException{
        checkingThread(true);
        BundleImpl impl = (BundleImpl) Atlas.getInstance().getBundle(dependencyBundle);
        if(impl==null){
            BundleInstallerFetcher.obtainInstaller().installTransitivelySync(new String[]{dependencyBundle});
        }
        impl = (BundleImpl) Atlas.getInstance().getBundle(dependencyBundle);
        if(impl==null){
            throw new BundleException("failed install deppendencyBundle : " +dependencyBundle);
        }else{
            requestRuntimeDependency(source,impl.getClassLoader(),resourceDependencyNeed);
        }
    }

    /**
     * 自定义异步安装时的界面提醒
     * @param reminder
     */
    public void setExternalBundleInstallReminder(ExternalBundleInstallReminder reminder){
        RuntimeVariables.sReminder = reminder;
    }

    /**
     * 设置bundle安装的校验器，会在bundle启动时回调提示校验，校验不通过将无法运行
     * @param checker
     */
    public void setBundleSecurityChecker(BundleVerifier checker){
        RuntimeVariables.sBundleVerifier = checker;
    }

    public void forceStopSelf(){
//        setPackageStoppedState
    }

    /*************************************************↑↑↑↑↑↑for public use↑↑↑↑↑↑*************************************************************
     **************************************************↑↑↑↑↑↑for public use↑↑↑↑↑↑*************************************************************
     **************************************************↑↑↑↑↑↑for public use↑↑↑↑↑↑***************************************************************/

    public static interface ExternalBundleInstallReminder{
        public Dialog createReminderDialog(Activity activity,String installingBundleName);
    }

    public static interface BundleVerifier{
        public boolean verifyBundle(String bundlePath);
    }
}
