/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.framework;

import android.taobao.atlas.bundleInfo.AtlasBundleInfoManager;
import android.taobao.atlas.framework.bundlestorage.BundleArchive;
import android.taobao.atlas.runtime.RuntimeVariables;
import android.taobao.atlas.runtime.DelegateResources;
import android.taobao.atlas.util.log.impl.AtlasMonitor;
import android.taobao.atlas.versionInfo.BaselineInfoManager;
import android.util.Log;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleException;
import org.osgi.framework.FrameworkEvent;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public final class BundleImpl implements Bundle {

    /**
     * the bundle location.
     */
    final String                    location;

    /**
     * the storage location.
     */
    File                      bundleDir;

    /**
     * the bundle archive file.
     */
    BundleArchive                         archive;

    /**
     * the bundle state.
     */
    int                             state;

    /**
     * the bundle classloader.
     */
    BundleClassLoader               classloader;
    boolean disabled = false;
    /**
     * create a new bundle object from InputStream. This is used when a new bundle is installed.
     *
     * @param location the bundle location.
     * @param stream the input stream.
     * @throws BundleException if something goes wrong.
     * @throws IOException
     */
    BundleImpl(final File bundleDir, final String location, final InputStream stream,
               final File file, String unique_tag, boolean installForCurrentVersion, long dexPatchVersion) throws BundleException, IOException{
        this.location = location;
        this.bundleDir = bundleDir;
        if(installForCurrentVersion) {
            Framework.notifyBundleListeners(BundleEvent.BEFORE_INSTALL, this);
        }
        if (stream != null) {
            this.archive = new BundleArchive(location,bundleDir, stream,unique_tag, dexPatchVersion);
        } else if (file != null) {
            this.archive = new BundleArchive(location,bundleDir, file,unique_tag, dexPatchVersion);
        }
        this.state = INSTALLED;
        if (installForCurrentVersion) {
            resolveBundle();
            Framework.bundles.put(location, this);
            // notify the listeners
            Framework.notifyBundleListeners(BundleEvent.INSTALLED, this);
        }
    }

    /**
     * reload a new bundle object from a storage location. Used after framework restarts.
     *
     * @param bcontext the bundle context.
     * @throws Exception if something goes wrong.
     */
    BundleImpl(final BundleContext bcontext) throws Exception{
        long start = System.currentTimeMillis();
        this.location = bcontext.location;
        long dexPatchVersion = BaselineInfoManager.instance().getDexPatchBundleVersion(location);
        Framework.notifyBundleListeners(BundleEvent.BEFORE_INSTALL, this);
        this.state = Bundle.INSTALLED;
        try {
            if(dexPatchVersion>0){
                try {
                    bundleDir = bcontext.dexPatchDir;
                    this.archive = new BundleArchive(location, bundleDir, bcontext.bundle_tag, dexPatchVersion);
                }catch(Throwable e){
                    bundleDir = bcontext.bundleDir;
                    this.archive = new BundleArchive(location, bundleDir, bcontext.bundle_tag, -1);
                }
            }else {
                bundleDir = bcontext.bundleDir;
                this.archive = new BundleArchive(location, bundleDir, bcontext.bundle_tag,-1);
            }
        } catch (Exception e) {
            if(e instanceof BundleArchive.MisMatchException){
                this.archive = null;
                BaselineInfoManager.instance().rollbackHardly();
                Map<String, Object> detail = new HashMap<>();
                detail.put("BundleImpl", "BundleImpl create failed!");
                AtlasMonitor.getInstance().report(AtlasMonitor.DD_BUNDLE_MISMATCH, detail, e);
                throw e;
            }else {
                throw new BundleException("Could not load bundle " + location, e.getCause());
            }
        }

        resolveBundle();
        Framework.bundles.put(location, this);

        // notify the listeners
        Framework.notifyBundleListeners(BundleEvent.INSTALLED, this);

        if (Framework.DEBUG_BUNDLES) {
            Log.i("Framework"," Bundle " + toString() + " loaded. " + (System.currentTimeMillis() - start) + " ms");
        }

    }

    private synchronized void resolveBundle() throws BundleException {

        if (this.archive == null) {
            throw new BundleException("Not a valid bundle: " + location);
        }

    	if (this.state == RESOLVED){
    		return;
    	}

    	if ( this.classloader == null){
	        // create the bundle classloader
            List<String> dependencies = AtlasBundleInfoManager.instance().getDependencyForBundle(location);
            String nativeLibDir = getArchive().getCurrentRevision().getRevisionDir().getAbsolutePath()+"/lib"+":"
                    + RuntimeVariables.androidApplication.getApplicationInfo().nativeLibraryDir+":"
                    +System.getProperty("java.library.path");
            if(dependencies!=null) {
                for (String str : dependencies) {
                    BundleImpl impl = (BundleImpl) Atlas.getInstance().getBundle(str);
                    if (impl != null) {
                        nativeLibDir += ":";
                        File dependencyLibDir = new File(impl.getArchive().getCurrentRevision().getRevisionDir(), "lib");
                        nativeLibDir += dependencyLibDir;
                    }
                }
            }
	        this.classloader = new BundleClassLoader(this,dependencies,nativeLibDir);
    	}
        state = RESOLVED;
        // notify the listeners
        Framework.notifyBundleListeners(0 /*LOADED*/, this);
    }

    /**
     * get the bundle id.
     *
     * @return the bundle id.
     * @see org.osgi.framework.Bundle#getBundleId()
     * @category Bundle
     */
    public long getBundleId() {
        return 0;
    }

    /**
     * get the bundle location.
     *
     * @return the bundle location.
     * @see org.osgi.framework.Bundle#getLocation()
     * @category Bundle
     */
    public String getLocation() {
        return location;
    }

    public BundleArchive getArchive() {
        return archive;
    }

    public ClassLoader getClassLoader() {
        return classloader;
    }

    /**
     * get a resource.
     *
     * @param name the name of the resource.
     * @return the URL.
     * @see org.osgi.framework.Bundle#getResource(java.lang.String)
     * @category Bundle
     */
    public URL getResource(final String name) {
        if (state == UNINSTALLED) {
            throw new IllegalStateException("Bundle " + toString() + " has been uninstalled");
        }
        return classloader.getResource(name);
    }

    /**
     * get the state of the bundle.
     *
     * @return the state.
     * @see org.osgi.framework.Bundle#getState()
     * @category Bundle
     */
    public int getState() {
        return state;
    }

    /**
     * start the bundle.
     *
     * @throws BundleException if the bundle cannot be resolved or the Activator throws an exception.
     * @see org.osgi.framework.Bundle#start()
     * @category Bundle
     */
    public void start() throws BundleException {
        startBundle();
    }

    @Override
    public void stop() throws BundleException {
        throw new IllegalStateException("Cannot stop bundle now");
    }

    /**
     * the actual starting happens here.
     *
     * @throws BundleException if the bundle cannot be resolved or the Activator throws an exception.
     */
    public synchronized void startBundle() {
        if (state == UNINSTALLED) {
            throw new IllegalStateException("Cannot start uninstalled bundle " + toString());
        }
        if (state == ACTIVE) {
            return;
        }
        if (state == INSTALLED) {
            throw new RuntimeException("can not start bundle which is not resolved");
        }
        state = STARTING;
        Framework.notifyBundleListeners(BundleEvent.BEFORE_STARTED, this);
        Framework.notifyBundleListeners(BundleEvent.STARTED, this);
        if (Framework.DEBUG_BUNDLES) {
            Log.i("Framework","Bundle " + toString() + " started.");
        }


    }

    public void setActive(){
        state = ACTIVE;
    }

    public boolean checkValidate(){
        boolean result = true;
        //check class
        if(classloader!=null){
            long time = System.currentTimeMillis();
            result = classloader.validateClasses();
            if(!result){
                return false;
            }

            if(!checkResources()){
                return false;
            }
        }
        return true;

    }

    public boolean checkResources(){
        if(!DelegateResources.checkAsset(getArchive().getArchiveFile().getAbsolutePath())){
            return false;
        }

        List<String> dependencies = AtlasBundleInfoManager.instance().getBundleInfo(location).getTotalDependency();
        for(String bundleName : dependencies){
            BundleImpl dependencyBundle = (BundleImpl)Atlas.getInstance().getBundle(bundleName);
            if(dependencyBundle==null || dependencyBundle.getArchive()==null ||
                    !DelegateResources.checkAsset(dependencyBundle.getArchive().getArchiveFile().getAbsolutePath())){
                return false;
            }
        }

        return true;
    }


    public boolean isDisabled(){
        return disabled;
    }

    /**
     * uninstall the bundle.
     * 
     * @throws BundleException if bundle is already uninstalled
     * @see org.osgi.framework.Bundle#uninstall()
     * @category Bundle
     */
    public synchronized void uninstall() throws BundleException {

        if (state == UNINSTALLED) {
            throw new IllegalStateException("Bundle " + toString() + " is already uninstalled.");
        }
        if (state == ACTIVE) {
            try {
                stop();
            } catch (Throwable t) {
                Framework.notifyFrameworkListeners(FrameworkEvent.ERROR, this, t);
            }
        }

        state = UNINSTALLED;

        classloader.cleanup(true);
        classloader = null;

        Framework.bundles.remove(getLocation());
        Framework.notifyBundleListeners(BundleEvent.UNINSTALLED, this);
    }
    
    public synchronized void optDexFile() {
    	this.getArchive().optDexFile();
    }

    /**
     * get a string representation of the bundle.
     * 
     * @return the string.
     * @category Object
     */
    public String toString() {
        return location;
    }
}
