/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.runtime.newcomponent;

import android.content.ComponentName;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.ActivityInfo;
import android.content.pm.ApplicationInfo;
import android.content.pm.ComponentInfo;
import android.content.pm.ProviderInfo;
import android.content.pm.ResolveInfo;
import android.content.pm.ServiceInfo;
import android.content.res.AssetManager;
import android.content.res.Resources;
import android.content.res.XmlResourceParser;
import android.os.Build;
import android.os.Bundle;
import android.taobao.atlas.hack.AtlasHacks;
import android.taobao.atlas.hack.Hack;
import android.taobao.atlas.runtime.RuntimeVariables;
import android.taobao.atlas.versionInfo.BaselineInfoManager;
import android.text.TextUtils;
import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * Created by guanjie on 15/3/5.
 */
public class AdditionalPackageManager {

    private static AdditionalPackageManager sBundlePackagemanager;
    private static int retryCount = 2;

    private AdditionalComponentIntentResolver mExternalActivity;
    private AdditionalComponentIntentResolver mExternalServices;
    private AdditionalComponentIntentResolver mExternalReceivers;
    private AdditionalComponentIntentResolver mExternalProviders;
    private final HashMap<IntentFilter,String> mReceiverIntentFilters = new HashMap<>();
    private final HashMap<String, Object> mProvidersByAuthority = new HashMap<>();

    public static synchronized AdditionalPackageManager getInstance(){
        if(sBundlePackagemanager == null){
            sBundlePackagemanager = new AdditionalPackageManager();
        }
        return sBundlePackagemanager;
    }

    public AdditionalPackageManager(){
        if(BaselineInfoManager.instance().isUpdated("com.taobao.maindex")) {
            boolean needRetry ;
            do {
                try {
                    retryCount--;
                    needRetry = false;
                    Class KernalBundleClass = RuntimeVariables.getRawClassLoader().loadClass("android.taobao.atlas.startup.patch.KernalBundle");
                    Field kernalBundleField = KernalBundleClass.getDeclaredField("kernalBundle");
                    Object kernalBundle = kernalBundleField.get(KernalBundleClass);
                    if (kernalBundle != null) {
                        Method method = KernalBundleClass.getDeclaredMethod("getRevisionZip");
                        method.setAccessible(true);
                        File file = (File) method.invoke(kernalBundle);
                        if (file != null) {
                            parseBundle(file.getAbsolutePath());
                        }
                    }
                } catch (Throwable e) {
                    needRetry = true;
                    e.printStackTrace();
                }
            }while(retryCount>0 && needRetry);
        }
    }

    /**
     * 解析变化 Manifest
     * @param zipPath
     * @return
     * @throws InvocationTargetException
     */
    private void parseBundle(String zipPath) throws InvocationTargetException{

        Object packageObj = null;
        Object pkgParser  = null;
        if(Build.VERSION.SDK_INT<=20) {
            pkgParser = AtlasHacks.PackageParser_constructor.getInstance(zipPath);
        }else{
            pkgParser = AtlasHacks.PackageParser_constructor.getInstance();
        }
        try {
            AssetManager assmgr = AssetManager.class.newInstance();
            AtlasHacks.AssetManager_addAssetPath.invoke(assmgr, RuntimeVariables.androidApplication.getApplicationInfo().sourceDir);
            int cookie = (Integer) AtlasHacks.AssetManager_addAssetPath.invoke(assmgr,zipPath);
            Resources res = new Resources(assmgr,RuntimeVariables.delegateResources.getDisplayMetrics(),RuntimeVariables.delegateResources.getConfiguration());
            XmlResourceParser parser = assmgr.openXmlResourceParser(cookie, "AndroidManifest.xml");
            packageObj = parsePackage(pkgParser,res,parser);
        } catch (Exception e) {
            e.printStackTrace();
            return;
        }
        if (packageObj == null) {
            return ;
        }
        AtlasHacks.PackageParser$Package_packageName.set(packageObj,RuntimeVariables.androidApplication.getPackageName());
        ApplicationInfo info = AtlasHacks.PackageParser$Package_applicationInfo.get(packageObj);
        info.name = RuntimeVariables.androidApplication.getApplicationInfo().name;
        info.className = RuntimeVariables.androidApplication.getApplicationInfo().className;
        info.taskAffinity = RuntimeVariables.androidApplication.getApplicationInfo().taskAffinity;
        info.permission = RuntimeVariables.androidApplication.getApplicationInfo().permission;
        info.processName = RuntimeVariables.androidApplication.getApplicationInfo().processName;
        info.theme = RuntimeVariables.androidApplication.getApplicationInfo().theme;
        info.flags = RuntimeVariables.androidApplication.getApplicationInfo().flags;
        info.uiOptions = RuntimeVariables.androidApplication.getApplicationInfo().uiOptions;
        info.backupAgentName = RuntimeVariables.androidApplication.getApplicationInfo().backupAgentName;
        info.descriptionRes = RuntimeVariables.androidApplication.getApplicationInfo().descriptionRes;
        info.targetSdkVersion = RuntimeVariables.androidApplication.getApplicationInfo().targetSdkVersion;
        info.compatibleWidthLimitDp = RuntimeVariables.androidApplication.getApplicationInfo().compatibleWidthLimitDp;
        info.uid = RuntimeVariables.androidApplication.getApplicationInfo().uid;
        info.largestWidthLimitDp = RuntimeVariables.androidApplication.getApplicationInfo().largestWidthLimitDp;
        info.enabled = RuntimeVariables.androidApplication.getApplicationInfo().enabled;
        info.requiresSmallestWidthDp = RuntimeVariables.androidApplication.getApplicationInfo().requiresSmallestWidthDp;
        info.packageName = RuntimeVariables.androidApplication.getApplicationInfo().packageName;

        AdditionalComponentIntentResolver activityResolver = new AdditionalComponentIntentResolver(AdditionalComponentIntentResolver.TYPE_ACTIVITY);
        AdditionalComponentIntentResolver serviceResolver  = new AdditionalComponentIntentResolver(AdditionalComponentIntentResolver.TYPE_SERVICE);
        AdditionalComponentIntentResolver receiverResolver  = new AdditionalComponentIntentResolver(AdditionalComponentIntentResolver.TYPE_ACTIVITY);
        AdditionalComponentIntentResolver providerResolver  = new AdditionalComponentIntentResolver(AdditionalComponentIntentResolver.TYPE_PROVIDER);


        ArrayList<Object> activityList = AtlasHacks.PackageParser$Package_activities.get(packageObj);
        ArrayList<Object> serviceList =  AtlasHacks.PackageParser$Package_services.get(packageObj);
        ArrayList<Object> providerList = AtlasHacks.PackageParser$Package_providers.get(packageObj);
        ArrayList<Object> receiverList = AtlasHacks.PackageParser$Package_receivers.get(packageObj);


        /**
         * activity
         */
        if (activityList != null) {
            for (Object activity : activityList) {
                try {
                    ActivityInfo activityInfo = (ActivityInfo) activity.getClass().getField("info").get(activity);
                    if(activityInfo.targetActivity!=null) {
                        activityInfo.taskAffinity = RuntimeVariables.androidApplication.getApplicationInfo().taskAffinity;
                    }
                    activityInfo.processName = TextUtils.isEmpty(activityInfo.processName) ? RuntimeVariables.androidApplication.getPackageName() : activityInfo.processName;
                    activityInfo.packageName = RuntimeVariables.androidApplication.getApplicationInfo().packageName;
                    activityResolver.addComponent(activity);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            if(activityResolver!=null && activityResolver.mComponents.size()>0) {
                setNewActivityResolver(activityResolver);
            }
        }

        /**
         * service
         */
        if (serviceList != null) {
            for (Object service : serviceList) {
                try {
                    ServiceInfo serviceInfo = (ServiceInfo) service.getClass().getField("info").get(service);
                    serviceInfo.packageName = RuntimeVariables.androidApplication.getApplicationInfo().packageName;
                    serviceInfo.processName = serviceInfo.processName==null ? RuntimeVariables.androidApplication.getPackageName() : serviceInfo.processName;
                    serviceResolver.addComponent(service);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            setNewServiceResolver(serviceResolver);
        }

        /**
         * provider
         */
        if(providerList!=null){
            for(Object provider : providerList){
                provider = providerList.get(0);
                try {
                    ProviderInfo providerInfo = AtlasHacks.PackageParser$Provider_info.get(provider);
                    providerInfo.processName = providerInfo.processName==null ? RuntimeVariables.androidApplication.getPackageName() : providerInfo.processName;
                    providerResolver.addComponent(provider);
                    if (providerInfo.authority != null) {
                        String names[] = providerInfo.authority.split(";");
                        providerInfo.authority = null;
                        for (int j = 0; j < names.length; j++) {
                            if (!mProvidersByAuthority.containsKey(names[j])) {
                                mProvidersByAuthority.put(names[j], provider);
                                if (providerInfo.authority == null) {
                                    providerInfo.authority = names[j];
                                } else {
                                    providerInfo.authority = providerInfo.authority + ";" + names[j];
                                }
                            }
                        }
                    }
                } catch (Throwable e) {
                    e.printStackTrace();
                }
            }
            setNewProviderResolver(providerResolver);
        }

        /**
         * receiver
         */
        if (receiverList != null) {
            for (Object activity : receiverList) {
                try {
                    ActivityInfo activityInfo = (ActivityInfo) activity.getClass().getField("info").get(activity);
                    receiverResolver.addComponent(activity);
                    activityInfo.processName = activityInfo.processName==null ? RuntimeVariables.androidApplication.getPackageName() : activityInfo.processName;
                    ArrayList<IntentFilter> intentFilters = (ArrayList<IntentFilter>) activity.getClass().getSuperclass().getDeclaredField("intents").get(activity);
                    if(intentFilters!=null){
                        for(IntentFilter filter : intentFilters){
                            mReceiverIntentFilters.put(filter,activityInfo.name);
                        }
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            if(receiverResolver!=null && receiverResolver.mComponents.size()>0) {
                setNewReceiverResolver(receiverResolver);
            }
        }
    }

    private static Object parsePackage(Object packageParserObj,Resources resources,XmlResourceParser parser){
        Hack.HackedMethod PackageParser_parsePackage = null;
        final String[] outError = new String[1];
        try {
            if(Build.VERSION.SDK_INT>=26){
                PackageParser_parsePackage = AtlasHacks.PackageParser.method("parseBaseApk",String.class,Resources.class, XmlResourceParser.class, int.class, String[].class);
                return PackageParser_parsePackage.invoke(packageParserObj, RuntimeVariables.androidApplication.getPackageName(),resources, parser, 0, outError);
            }else if (Build.VERSION.SDK_INT > 20) {
                PackageParser_parsePackage = AtlasHacks.PackageParser.method("parseBaseApk", Resources.class, XmlResourceParser.class, int.class, String[].class);
                return PackageParser_parsePackage.invoke(packageParserObj, resources, parser, 0, outError);
            }else{
                try{
                    PackageParser_parsePackage = AtlasHacks.PackageParser.method("parsePackage", Resources.class, XmlResourceParser.class,int.class,String[].class);
                    return PackageParser_parsePackage.invoke(packageParserObj,resources,parser,0,outError);
                }catch(Hack.HackDeclaration.HackAssertionException e){
                    PackageParser_parsePackage = AtlasHacks.PackageParser.method("parsePackage", Resources.class, XmlResourceParser.class,int.class,boolean.class,String[].class);
                    return PackageParser_parsePackage.invoke(packageParserObj,resources,parser,0,false,outError);
                }
            }
        }catch(Throwable e){
            final ApplicationInfo app_info = RuntimeVariables.androidApplication.getApplicationInfo();
            if((app_info.flags & ApplicationInfo.FLAG_DEBUGGABLE) != 0){
                throw new RuntimeException(e);
            }
            return null;
        }
    }

    void setNewActivityResolver(AdditionalComponentIntentResolver resolver){
        mExternalActivity = resolver;
    }

    void setNewServiceResolver(AdditionalComponentIntentResolver resolver){
        mExternalServices = resolver;
    }

    void setNewReceiverResolver(AdditionalComponentIntentResolver resolver){
        mExternalReceivers = resolver;
    }

    void setNewProviderResolver(AdditionalComponentIntentResolver resolver){
        mExternalProviders = resolver;
    }

    private <T extends ComponentInfo> ResolveInfo queryIntentResolveInfo(Intent intent,Class<T> infoClass ){
        if(intent==null){
            return null;
        }
        AdditionalComponentIntentResolver resolver = null;
        if(infoClass == ActivityInfo.class){
            resolver = mExternalActivity;
        }else{
            resolver = mExternalServices;
        }
        if(resolver!=null) {
            ComponentName comp = intent.getComponent();
            if (comp == null) {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.ICE_CREAM_SANDWICH_MR1) {
                    if (intent.getSelector() != null) {
                        intent = intent.getSelector();
                        comp = intent.getComponent();
                    }
                }
            }
            if (comp != null) {
                Object obj = resolver.mComponents.get(comp);
                if (obj!=null){
                    try {
                        final ResolveInfo ri = new ExternalResolverInfo();
                        if(resolver == mExternalActivity) {
                            ri.activityInfo = (ActivityInfo) obj.getClass().getField("info").get(obj);
                        }else{
                            ri.serviceInfo = (ServiceInfo) obj.getClass().getField("info").get(obj);
                        }
                        return ri;
                    }catch(Exception e){
                        return null;
                    }
                }else{
                    return null;
                }
            }else{
                // 先检测包名
                if (!TextUtils.isEmpty(intent.getPackage()) && !TextUtils.equals(intent.getPackage(), RuntimeVariables.androidApplication.getPackageName())) {
                    return null;
                }
                List<ResolveInfo> list = resolver.queryIntent(intent,
                        intent.resolveTypeIfNeeded(RuntimeVariables.androidApplication.getContentResolver()), false);
                if(list!=null && list.size()>0){
                    return new ExternalResolverInfo(list.get(0));
                }else{
                    return null;
                }
            }
        }else{
            return null;
        }
    }

    public ProviderInfo resolveContentProvider(String name){
        final Object provider = mProvidersByAuthority.get(name);
        if(provider!=null){
            ProviderInfo pi = new ProviderInfo(AtlasHacks.PackageParser$Provider_info.get(provider));
            try {
                Field metaDataField = provider.getClass().getSuperclass().getDeclaredField("metaData");
                pi.metaData = (Bundle)metaDataField.get(provider);
            } catch (Throwable e) {
                e.printStackTrace();
            }
            return pi;
        }
        return null;

    }

    public List<ResolveInfo> queryIntentActivities(Intent intent){
        ResolveInfo info =  AdditionalPackageManager.getInstance().queryIntentResolveInfo(intent,ActivityInfo.class);
        if(info!=null){
            List<ResolveInfo> rf = new ArrayList<ResolveInfo>(1);
            rf.add(info);
            return rf;
        }
        return null;
    }

    public List<ResolveInfo> queryIntentService(Intent intent){
        ResolveInfo info =  AdditionalPackageManager.getInstance().queryIntentResolveInfo(intent,ServiceInfo.class);
        if(info!=null){
            List<ResolveInfo> rf = new ArrayList<ResolveInfo>(1);
            rf.add(info);
            return rf;
        }
        return null;
    }

    public <T extends ComponentInfo> T  getNewComponentInfo(ComponentName comonent,Class<T> ComponentInfoClass){
        AdditionalComponentIntentResolver resolver = null;
        if(ComponentInfoClass == ActivityInfo.class){
            resolver = mExternalActivity;
        }else if(ComponentInfoClass == ServiceInfo.class){
            resolver = mExternalServices;
        }else if(ComponentInfoClass== ProviderInfo.class){
            resolver = mExternalProviders;
        }
        if(resolver != null && resolver.mComponents!=null){
            Object obj = resolver.mComponents.get(comonent);
            if(obj!=null){
                try {
                    ComponentInfo info = (ComponentInfo) obj.getClass().getField("info").get(obj);
                    return (T)info;
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                } catch (NoSuchFieldException e) {
                    e.printStackTrace();
                }
            }
        }
        return null;
    }

    public ActivityInfo getReceiverInfo(ComponentName component){
        if(mExternalReceivers!=null && mExternalReceivers.mComponents!=null){
            Object obj = mExternalReceivers.mComponents.get(component);
            if(obj!=null){
                try {
                    ActivityInfo info = (ActivityInfo) obj.getClass().getField("info").get(obj);
                    return info;
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                } catch (NoSuchFieldException e) {
                    e.printStackTrace();
                }
            }
        }
        return null;
    }

    public List<ResolveInfo> queryIntentReceivers(Intent intent){
        if(intent==null){
            return null;
        }
        if(mExternalReceivers!=null) {
            ComponentName comp = intent.getComponent();
            if (comp == null) {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.ICE_CREAM_SANDWICH_MR1) {
                    if (intent.getSelector() != null) {
                        intent = intent.getSelector();
                        comp = intent.getComponent();
                    }
                }
            }
            if (comp != null) {
                Object obj = mExternalReceivers.mComponents.get(comp);
                if (obj!=null){
                    try {
                        List<ResolveInfo> list = new ArrayList<ResolveInfo>(1);
                        ResolveInfo info = new ResolveInfo();
                        info.activityInfo = (ActivityInfo) obj.getClass().getField("info").get(obj);
                        return list;
                    }catch(Exception e){
                        return null;
                    }
                }else{
                    return null;
                }
            }else{
                // 先检测包名
                if (!TextUtils.isEmpty(intent.getPackage()) && !TextUtils.equals(intent.getPackage(), RuntimeVariables.androidApplication.getPackageName())) {
                    return null;
                }
                return mExternalReceivers.queryIntent(intent,
                        intent.resolveTypeIfNeeded(RuntimeVariables.androidApplication.getContentResolver()), false);
            }
        }else{
            return null;
        }
    }

    public List<ProviderInfo> queryContentProviders(String processName){
        if(mExternalProviders!=null && mExternalProviders.mComponents!=null){
            List<ProviderInfo> infos = new ArrayList<>();
            try {
                for (Object provider : mExternalProviders.mComponents.values()) {
                    ProviderInfo info = (ProviderInfo) provider.getClass().getField("info").get(provider);
                    info.processName = TextUtils.isEmpty(info.processName) ? RuntimeVariables.androidApplication.getPackageName() : info.processName;
                    if(info!=null){
                        if(processName==null || processName.equals(info.processName)){
                            infos.add(info);
                        }
                    }
                }
                return infos;
            }catch(Throwable e){
                e.printStackTrace();
                return null;
            }
        }
        return null;
    }

    public IntentFilter getAdditionIntentFilter(){
        if(mReceiverIntentFilters != null){
            IntentFilter allAdditionalFilter = new IntentFilter();
            try {
                Field mActionsField = IntentFilter.class.getDeclaredField("mActions");
                mActionsField.setAccessible(true);
                for (IntentFilter filter : mReceiverIntentFilters.keySet()) {
                    ArrayList<String> mActions = (ArrayList<String>) mActionsField.get(filter);
                    if(mActions!=null){
                        for(String action : mActions){
                            allAdditionalFilter.addAction(action);
                        }
                    }
                }
            }catch(Throwable e){
                e.printStackTrace();
            }
            return allAdditionalFilter;
        }
        return null;
    }

    public static class ExternalResolverInfo extends ResolveInfo{
        public ExternalResolverInfo(){
        }

        public ExternalResolverInfo(ResolveInfo info){
            super(info);
        }
    }

}
