/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.startup;

import android.annotation.TargetApi;
import android.app.ActivityManager;
import android.app.Application;
import android.app.Notification;
import android.app.NotificationManager;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.res.Resources;
import android.os.Build;
import android.os.Process;
import android.taobao.atlas.startup.patch.KernalBundle;
import android.taobao.atlas.startup.patch.KernalConstants;
import android.text.TextUtils;
import android.util.Log;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;

/**
 * Created by guanjie on 16/10/20.
 */

/**
 * android.taobao.atlas.start.* package can't be tpatched
 * any class in this package could not import external classes except framework classes
 *
 *  * android.taobao.atlas.start.* package can't be tpatched
 * any class in this package could not import external classes except framework classes
 *
 *  * android.taobao.atlas.start.* package can't be tpatched
 * any class in this package could not import external classes except framework classes
 */
public class AtlasBridgeApplication extends Application{

    public  Object mBridgeApplicationDelegate;
    @Override
    protected void attachBaseContext(Context base) {
        super.attachBaseContext(base);
        if (!isApplicationNormalCreate(base)) {
            android.os.Process.killProcess(android.os.Process.myPid());
        }
        // *0 checkload kernalpatch
        boolean isUpdated = isUpdated(getBaseContext());
        KernalConstants.baseContext = getBaseContext();
        KernalConstants.APK_PATH = getBaseContext().getApplicationInfo().sourceDir;
        KernalConstants.RAW_APPLICATION_NAME = getClass().getName();
        DexLoadBooster dexBooster = new DexLoadBooster();
        dexBooster.init(getBaseContext());
        KernalConstants.dexBooster = dexBooster;
        boolean hasKernalPatched  = false;
        boolean isMainProcess = getBaseContext().getPackageName().equals(KernalConstants.PROCESS);
        if(isUpdated){
            if (!isMainProcess) {
                android.os.Process.killProcess(android.os.Process.myPid());
            }
            File storageDir = new File(getFilesDir(),"storage");
            File bundleBaseline = new File(getFilesDir(),"bundleBaseline");
            deleteDirectory(storageDir);
            KernalVersionManager.instance().removeBaseLineInfo();
            if(storageDir.exists() || bundleBaseline.exists()){
                android.os.Process.killProcess(Process.myPid());
            }
            storePackageVersion(base);
            KernalVersionManager.instance().init();
            System.setProperty("APK_INSTALLED", "true");
        }else{
            if(KernalConstants.PROCESS.contains(":dex2oat")){
                return;
            }
            KernalVersionManager.instance().init();
            if(!KernalBundle.checkLoadKernalDebugPatch(this)){
                if(KernalBundle.hasKernalPatch()) {
                    //has patch ? true -> must load successed
                    hasKernalPatched = KernalBundle.checkloadKernalBundle(this, KernalConstants.PROCESS);
                    if (!hasKernalPatched) {
                        // load failed
                        if(isMainProcess) {
                            KernalVersionManager.instance().rollbackHardly();
                        }
                        android.os.Process.killProcess(android.os.Process.myPid());
                    }
                }else{
                    //remove deprecated info
                    if(isMainProcess) {
                        KernalBundle.clear();
                    }
                }
            }
        }

        try {
            //初始化baselineinfomanager
            Class BaselineInfoManagerClazz = getBaseContext().getClassLoader().loadClass("android.taobao.atlas.versionInfo.BaselineInfoManager");
            Method instanceMethod = BaselineInfoManagerClazz.getDeclaredMethod("instance");
            Object instance = instanceMethod.invoke(BaselineInfoManagerClazz);
            Field mVersionManager = BaselineInfoManagerClazz.getDeclaredField("mVersionManager");
            mVersionManager.setAccessible(true);
            mVersionManager.set(instance,KernalVersionManager.instance());

            Class BridgeApplicationDelegateClazz = getBaseContext().getClassLoader().loadClass("android.taobao.atlas.bridge.BridgeApplicationDelegate");
            Class<?>[] parTypes=new Class<?>[8];
            parTypes[0]= Application.class;
            parTypes[1]= String.class;
            parTypes[2]= String.class;
            parTypes[3]= long.class;
            parTypes[4]= long.class;
            parTypes[5]= String.class;
            parTypes[6]= boolean.class;
            parTypes[7]= Object.class;
            Constructor<?> con = BridgeApplicationDelegateClazz.getConstructor(parTypes);
            mBridgeApplicationDelegate = con.newInstance(this,KernalConstants.PROCESS,KernalConstants.INSTALLED_VERSIONNAME,
                    KernalConstants.INSTALLED_VERSIONCODE,KernalConstants.LASTUPDATETIME,KernalConstants.APK_PATH,isUpdated,KernalConstants.dexBooster);
            Method method = BridgeApplicationDelegateClazz.getDeclaredMethod("attachBaseContext");
            method.invoke(mBridgeApplicationDelegate);
        } catch (Throwable e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public Resources getResources() {
        Resources rs =  super.getResources();
        if (rs == null){
            PackageManager pm = getPackageManager();
            try {
                rs = pm.getResourcesForApplication(getApplicationInfo());
            } catch (PackageManager.NameNotFoundException e) {
                e.printStackTrace();
                return null;
            }
        }
        return rs;
    }

    @Override
    public void onCreate() {
        super.onCreate();
        if(!KernalConstants.PROCESS.contains(":dex2oat")){
            try {
                Method method = mBridgeApplicationDelegate.getClass().getDeclaredMethod("onCreate");
                method.invoke(mBridgeApplicationDelegate);
            } catch (Throwable e) {
                e.printStackTrace();
            }
        }
    }

    private boolean isApplicationNormalCreate(Context base) {
        Throwable ex = new Throwable();
        StackTraceElement[] stackElements = ex.getStackTrace();
        if (stackElements != null) {
            for (int i = 0; i < stackElements.length; i++) {
                if (stackElements[i].getClassName().equalsIgnoreCase("android.taobao.atlas.runtime.InstrumentationHook")) {
                    return false;
                }
            }
        }

        ApplicationInfo info = base.getApplicationInfo();
        if (info == null) {
            return false;
        }
        String apkPath = info.sourceDir;
        String nativeLibDir = info.nativeLibraryDir;
        if (apkPath == null || !new File(apkPath).exists()) {
            checkShowErrorNotification("InvalidApkPath");
            return false;
        }

        if (nativeLibDir == null || !new File(nativeLibDir).exists()) {
            checkShowErrorNotification("InvalidLibPath");
            return false;
        }

        int pid = android.os.Process.myPid();
        try {
            ActivityManager mActivityManager = (ActivityManager) base.getSystemService(Context.ACTIVITY_SERVICE);
            for (ActivityManager.RunningAppProcessInfo appProcess : mActivityManager.getRunningAppProcesses()) {
                if (appProcess.pid == pid) {
                    KernalConstants.PROCESS =  appProcess.processName;
                }
            }
        } catch (Exception e) {
        }
        if(TextUtils.isEmpty(KernalConstants.PROCESS)){
            Log.e("AtlasBridgeApplication","getProcess failed");
            return false;
        }
        return true;
    }

    /**
     * TODO 如果supportv4 在子dex这种写法会存在问题
     *
     * @param errorInfo
     */
    @TargetApi(Build.VERSION_CODES.JELLY_BEAN)
    private void checkShowErrorNotification(String errorInfo) {
        SharedPreferences preferences = getSharedPreferences("err_log", Context.MODE_PRIVATE);
        int errorCount = preferences.getInt(errorInfo, 0);
        if (errorCount >= 3) {
            Notification.Builder mBuilder =
                    new Notification.Builder(this)
                            .setSmallIcon(this.getResources().getIdentifier("icon","drawable",getPackageName()))
                            .setContentTitle("提示").setAutoCancel(true)
                            .setContentText("应用安装不完整,请您卸载重新安装!");

            NotificationManager mNotificationManager =
                    (NotificationManager) getSystemService(Context.NOTIFICATION_SERVICE);
            mNotificationManager.notify(123, mBuilder.build());
        } else {
            preferences.edit().putInt(errorInfo, ++errorCount).commit();
        }
    }

    public  boolean isUpdated(Context context){
        PackageInfo packageInfo = null;
        // 获取当前的版本号
        try {
            PackageManager packageManager = context.getPackageManager();
            packageInfo = packageManager.getPackageInfo(context.getPackageName(), 0);
        } catch (Exception e) {
            // 不可能发生
            android.os.Process.killProcess(android.os.Process.myPid());
        }
        KernalConstants.INSTALLED_VERSIONNAME = packageInfo.versionName;
        KernalConstants.INSTALLED_VERSIONCODE = packageInfo.versionCode;
        KernalConstants.LASTUPDATETIME = packageInfo.lastUpdateTime;
        if(TextUtils.isEmpty(KernalConstants.INSTALLED_VERSIONNAME )){
            //不可能发生
            Log.e("AtlasBridgeApplication","version name is empty ");
            android.os.Process.killProcess(android.os.Process.myPid());
        }

        File metafile = new File(context.getFilesDir(), "storage/version_meta");
        if (metafile.exists()) {
            try {
                DataInputStream in = new DataInputStream(new FileInputStream(metafile));
                String storedVersionName = in.readUTF();
                long   storedVersionCode = in.readLong();
                long   storedLastUpdateTime = in.readLong();
                String storedApkPath     = in.readUTF();

                System.setProperty("APP_VERSION_TAG",KernalConstants.INSTALLED_VERSIONNAME);
                // 检测之前的版本记录
                if(packageInfo.versionCode == storedVersionCode &&
                        TextUtils.equals(packageInfo.versionName, storedVersionName) &&
                        packageInfo.lastUpdateTime == storedLastUpdateTime &&
                        context.getApplicationInfo().sourceDir.equals(storedApkPath) &&
                        !needRollback()){
                    return false;
                }
            }catch(Throwable e){
                throw new RuntimeException(e);
            }
        }
        return true;
    }

    private void storePackageVersion(Context base){
        File file = new File(base.getFilesDir(), "storage/version_meta");
        DataOutputStream out = null;
        try {
            if (!file.getParentFile().exists()) {
                file.getParentFile().mkdirs();
            }

            FileOutputStream fos = new FileOutputStream(file);
            out = new DataOutputStream(fos);
            out.writeUTF(KernalConstants.INSTALLED_VERSIONNAME);
            out.writeLong(KernalConstants.INSTALLED_VERSIONCODE);
            out.writeLong(KernalConstants.LASTUPDATETIME);
            out.writeUTF(KernalConstants.APK_PATH);
            out.flush();
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            try{
                out.close();
            }catch (Throwable e){}
        }
    }

    public boolean needRollback(){
        File baseLineDir = new File(getBaseContext().getFilesDir().getAbsolutePath() + File.separatorChar + "bundleBaseline");
        if(baseLineDir.exists()){
            File deprecated = new File(baseLineDir, "deprecated_mark");
            if(deprecated.exists()){
                return true;
            }
        }
        return false;
    }

    public void deleteDirectory(final File path) {
        if(!path.exists()){
            return;
        }
        final File[] files = path.listFiles();
        if (files == null){
            return;
        }
        for (int i = 0; i < files.length; i++) {
            if (files[i].isDirectory()) {
                deleteDirectory(files[i]);
            } else {
                files[i].delete();
            }
        }
        path.delete();
    }
}
