/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */
package com.taobao.android.builder.tools.command;

import org.codehaus.plexus.util.cli.CommandLineException;
import org.codehaus.plexus.util.cli.CommandLineUtils;
import org.codehaus.plexus.util.cli.Commandline;
import org.codehaus.plexus.util.cli.StreamConsumer;
import org.codehaus.plexus.util.cli.shell.Shell;
import org.gradle.api.logging.Logger;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *
 */
public interface CommandExecutor {
    /**
     * Sets the plexus logger.
     *
     * @param logger the plexus logger
     */
    void setLogger(Logger logger);

    /**
     * Executes the command for the specified executable and list of command options.
     *
     * @param executable the name of the executable (csc, xsd, etc).
     * @param commands   the command options for the compiler/executable
     * @throws ExecutionException if compiler or executable writes anything to the standard error stream or if the process returns a
     *                            process result != 0.
     */
    void executeCommand(String executable, List<String> commands) throws ExecutionException;

    /**
     * Executes the command for the specified executable and list of command options.
     *
     * @param executable         the name of the executable (csc, xsd, etc).
     * @param commands           the commands options for the compiler/executable
     * @param failsOnErrorOutput if true, throws an <code>ExecutionException</code> if there the compiler or executable writes anything
     *                           to the error output stream. By default, this value is true
     * @throws ExecutionException if compiler or executable writes anything to the standard error stream (provided the
     *                            failsOnErrorOutput is not false) or if the process returns a process result != 0.
     */
    void executeCommand(String executable, List<String> commands, boolean failsOnErrorOutput)
            throws ExecutionException;

    /**
     * Executes the command for the specified executable and list of command options. If the compiler or executable is
     * not within the environmental path, you should use this method to specify the working directory. Always use this
     * method for executables located within the local maven repository.
     *
     * @param executable       the name of the executable (csc, xsd, etc).
     * @param commands         the command options for the compiler/executable
     * @param workingDirectory the directory where the command will be executed
     * @throws ExecutionException if compiler or executable writes anything to the standard error stream (provided the
     *                            failsOnErrorOutput is not false) or if the process returns a process result != 0.
     */
    void executeCommand(String executable, List<String> commands, File workingDirectory, boolean failsOnErrorOutput)
            throws ExecutionException;

    /**
     * Returns the process result of executing the command. Typically a value of 0 means that the process executed
     * successfully.
     *
     * @return the process result of executing the command
     */
    int getResult();

    /**
     * @return the process id for the executed command.
     */
    long getPid();

    /**
     * Returns the standard output from executing the command.
     *
     * @return the standard output from executing the command
     */
    String getStandardOut();

    /**
     * Returns the standard error from executing the command.
     *
     * @return the standard error from executing the command
     */
    String getStandardError();

    /**
     * Adds an environment variable with the specified name and value to the executor.
     */
    void addEnvironment(String name, String value);

    void setErrorListener(ErrorListener errorListener);

    void setCustomShell(Shell s);

    void setCaptureStdOut(boolean captureStdOut);

    void setCaptureStdErr(boolean captureStdErr);

    /**
     *
     */
    public interface ErrorListener {
        boolean isError(String error);
    }

    /**
     * Provides factory services for creating a default instance of the command executor.
     */
    class Factory {

        /**
         * Constructor
         */
        private Factory() {
        }

        private static final class DefaultCommandExecutor implements CommandExecutor {
            private Map<String, String> environment;
            /**
             * Instance of a plugin logger.
             */
            private Logger logger;
            /**
             * Standard Out
             */
            private StreamConsumer stdOut;
            /**
             * Standard Error
             */
            private ErrorStreamConsumer stdErr;
            /**
             * Process result
             */
            private int result;
            /*
             */
            private ErrorListener errorListener;
            long pid;
            private Commandline commandline;
            private Shell customShell;

            private boolean captureStdOut;
            private boolean captureStdErr;

            @Override
            public void setLogger(Logger logger) {
                this.logger = logger;
            }

            @Override
            public void executeCommand(String executable, List<String> commands) throws ExecutionException {
                executeCommand(executable, commands, null, true);
            }

            @Override
            public void executeCommand(String executable, List<String> commands, boolean failsOnErrorOutput)
                    throws ExecutionException {
                executeCommand(executable, commands, null, failsOnErrorOutput);
            }

            @Override
            public void executeCommand(String executable, List<String> commands, File workingDirectory,
                                       boolean failsOnErrorOutput) throws ExecutionException {
                if (commands == null) {
                    commands = new ArrayList<String>();
                }
                stdOut = new StreamConsumerImpl(logger, captureStdOut);
                stdErr = new ErrorStreamConsumer(logger, errorListener, captureStdErr);
                commandline = new Commandline();
                if (customShell != null) {
                    commandline.setShell(customShell);
                }
                commandline.setExecutable(executable);

                // Add the environment variables as needed
                if (environment != null) {
                    for (Map.Entry<String, String> entry : environment.entrySet()) {
                        commandline.addEnvironment(entry.getKey(), entry.getValue());
                    }
                }

                commandline.addArguments(commands.toArray(new String[commands.size()]));
                if (workingDirectory != null && workingDirectory.exists()) {
                    commandline.setWorkingDirectory(workingDirectory.getAbsolutePath());
                }
                try {
                    logger.debug("ANDROID-040-000: Executing command: Commandline = " + commandline);
                    result = CommandLineUtils.executeCommandLine(commandline, stdOut, stdErr);
                    if (logger != null) {
                        logger.debug("ANDROID-040-000: Executed command: Commandline = " + commandline + ", Result = "
                                + result);
                    } else {
                        System.out.println("ANDROID-040-000: Executed command: Commandline = " + commandline
                                + ", Result = " + result);
                    }
                    if (failsOnErrorOutput && stdErr.hasError() || result != 0) {
                        throw new ExecutionException("ANDROID-040-001: Could not execute: Command = "
                                + commandline.toString() + ", Result = " + result);
                    }
                } catch (CommandLineException e) {
                    throw new ExecutionException("ANDROID-040-002: Could not execute: Command = "
                            + commandline.toString() + ", Error message = " + e.getMessage());
                }
                setPid(commandline.getPid());
            }

            @Override
            public int getResult() {
                return result;
            }

            @Override
            public String getStandardOut() {
                if (!captureStdOut) {
                    throw new IllegalStateException("Unable to provide StdOut since it was not captured");
                }
                return stdOut.toString();
            }

            @Override
            public String getStandardError() {
                if (!captureStdErr) {
                    throw new IllegalStateException("Unable to provide StdOut since it was not captured");
                }
                return stdErr.toString();
            }

            @Override
            public void addEnvironment(String name, String value) {
                if (environment == null) {
                    environment = new HashMap<String, String>();
                }
                environment.put(name, value);
            }

            @Override
            public void setErrorListener(ErrorListener errorListener) {
                this.errorListener = errorListener;
            }

            public void setPid(long pid) {
                this.pid = pid;
            }

            @Override
            public long getPid() {
                return pid;
            }

            @Override
            public void setCustomShell(Shell shell) {
                this.customShell = shell;
            }

            @Override
            public void setCaptureStdOut(boolean captureStdOut) {
                this.captureStdOut = captureStdOut;
            }

            @Override
            public void setCaptureStdErr(boolean captureStdErr) {
                this.captureStdErr = captureStdErr;
            }
        }

        /**
         * StreamConsumer instance that buffers the entire output
         */
        static class StreamConsumerImpl implements StreamConsumer {
            private StringBuffer sb = new StringBuffer();
            private final Logger logger;
            private boolean captureStdOut;

            public StreamConsumerImpl(Logger logger, boolean captureStdOut) {
                this.logger = logger;
                this.captureStdOut = captureStdOut;
            }

            @Override
            public void consumeLine(String line) {
                if (captureStdOut) {
                    sb.append(line);
                }
                if (logger != null) {
                    logger.debug(line);
                }
            }

            /**
             * Returns the stream
             *
             * @return the stream
             */
            @Override
            public String toString() {
                return sb.toString();
            }
        }

        /**
         * Provides behavior for determining whether the command utility wrote anything to the Standard Error Stream.
         * NOTE: I am using this to decide whether to fail the NMaven build. If the compiler implementation chooses to
         * write warnings to the error stream, then the build will fail on warnings!!!
         */
        static class ErrorStreamConsumer implements StreamConsumer {
            /**
             * Is true if there was anything consumed from the stream, otherwise false
             */
            private boolean error;
            /**
             * Buffer to store the stream
             */
            private StringBuffer sbe = new StringBuffer();
            private final Logger logger;
            private final ErrorListener errorListener;
            private boolean captureStdErr;

            public ErrorStreamConsumer(Logger logger, ErrorListener errorListener, boolean captureStdErr) {
                this.logger = logger;
                this.errorListener = errorListener;
                this.captureStdErr = captureStdErr;

                if (logger == null) {
                    System.out.println("ANDROID-040-003: Error Log not set: Will not output error logs");
                }
                error = false;
            }

            @Override
            public void consumeLine(String line) {
                if (captureStdErr) {
                    sbe.append(line);
                }
                if (logger != null) {
                    logger.info(line);
                }
                if (errorListener != null) {
                    error = errorListener.isError(line);
                } else {
                    error = true;
                }
            }

            /**
             * Returns false if the command utility wrote to the Standard Error Stream, otherwise returns true.
             *
             * @return false if the command utility wrote to the Standard Error Stream, otherwise returns true.
             */
            public boolean hasError() {
                return error;
            }

            /**
             * Returns the error stream
             *
             * @return error stream
             */
            @Override
            public String toString() {
                return sbe.toString();
            }
        }

        /**
         * Returns a default instance of the command executor
         *
         * @return a default instance of the command executor
         */
        public static CommandExecutor createDefaultCommmandExecutor() {
            return new DefaultCommandExecutor();

        }
    }
}
