/*
 *
 *
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "[]"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2016 Alibaba Group
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 *
 *
 */
package com.taobao.android.object;
/*
 *
 *
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "[]"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2016 Alibaba Group
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 *
 *
 */

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.taobao.android.APatchTool;
import com.taobao.android.apatch.Build;
import com.taobao.android.apatch.utils.Formater;
import com.taobao.android.differ.dex.BundleDiffResult;
import com.taobao.android.differ.dex.PatchException;
import org.apache.commons.io.FileUtils;
import org.jf.dexlib2.dexbacked.DexBackedClassDef;
import org.jf.dexlib2.dexbacked.DexBackedField;
import org.jf.dexlib2.dexbacked.DexBackedMethod;
import org.jf.dexlib2.iface.ClassDef;
import org.jf.dexlib2.iface.reference.MethodReference;
import org.jf.dexlib2.util.ReferenceUtil;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.regex.Pattern;

/**
 * Created by shenghua.nish on 2016-03-13 上午10:02.
 */
public class DexDiffInfo {

    private static final Pattern p = Pattern.compile("Ljava/lang/|Ljava/util/|Landroid/app/|Landroid/util/");
    private Set<DexBackedClassDef> oldClasses = Collections.synchronizedSet(new HashSet<DexBackedClassDef>());
    private Set<DexBackedClassDef> newClasses = Collections.synchronizedSet(new HashSet<DexBackedClassDef>());
    public static Set<DexBackedClassDef> addedClasses = Collections.synchronizedSet(new HashSet<DexBackedClassDef>());
    public static Set<DexBackedClassDef> modifiedClasses = Collections.synchronizedSet(new HashSet<DexBackedClassDef>());
    private Set<DexBackedField> addedFields = Collections.synchronizedSet(new HashSet<DexBackedField>());
    private Set<DexBackedField> modifiedFields = Collections.synchronizedSet(new HashSet<DexBackedField>());
    public static Set<DexBackedMethod> addedMethods = Collections.synchronizedSet(new HashSet<DexBackedMethod>());
    public static Set<DexBackedMethod> modifiedMethods = Collections.synchronizedSet(new HashSet<DexBackedMethod>());
    private Map<String, ClassDiffInfo> classDiffInfoMap = new HashMap<String, ClassDiffInfo>();
    private Set<String> usedMethods = Collections.synchronizedSet(new HashSet<String>());

    public DexDiffInfo() {

    }

    public Map<String, ClassDiffInfo> getClassDiffInfoMap() {
        return classDiffInfoMap;
    }

    public Set<DexBackedClassDef> getModifiedClasses() {
        return modifiedClasses;
    }

    public void setAddedClasses(ClassDef classDef){
        if (!addedClasses.contains(classDef)){
            addedClasses.add((DexBackedClassDef) classDef);
        }
    }


    public Set<DexBackedClassDef> getAddedClasses(){
        return addedClasses;
    }

    public Set<DexBackedMethod> getModifiedMethods() {
        return modifiedMethods;
    }

    public Set<DexBackedMethod> getAddedMethods() {
        return addedMethods;
    }

    public static DexBackedClassDef getModifiedClasses(String clazz) {
        for (DexBackedClassDef classDef : modifiedClasses) {
            if (classDef.getType().equals(clazz)) {
                return classDef;
            }
        }
        return null;
    }


    public void addModifiedMethods(DexBackedMethod method) throws PatchException {
        System.out.println("add modified Method:" + method.getReturnType()
                + "  " + method.getName() + "("
                + Formater.formatStringList(method.getParameterTypes())
                + ")  in Class:" + method.getDefiningClass());
        this.modifiedMethods.add(method);

        if (!modifiedClasses.contains(method.classDef)) {
            modifiedClasses.add(method.classDef);
            String className = method.classDef.getType();
            addManifestModifiedClass(className);
        }
    }


    private void addManifestModifiedClass(String className) {
        Build.modifiedClasses.add(className.substring(1, className.length() - 1).replace('/', '.'));
        System.out.println("addModifiedClass: " + className.substring(1, className.length() - 1).replace('/', '.'));
    }


    public void addManifestAddClass(String className) {
        Build.addClasses.add(className.substring(1, className.length() - 1).replace('/', '.'));
        System.out.println("addClass: " + className.substring(1, className.length() - 1).replace('/', '.'));
    }


    public static void addUsedMethods(MethodReference methodReference) {
        String className = methodReference.getDefiningClass();
//        if (Build.addClasses.contains(className.substring(1, className.length() - 1).replace('/', '.'))){
//            return;
//        }
        String method = methodReference.getName();
        String params;
        if (APatchTool.mappingMap == null) {
            params = methodReference.getParameterTypes().toString().replace(',', '|').replaceAll(" ", "");
        } else {
            params = getParamsType(methodReference.getParameterTypes());
        }


        System.out.println("add used method:" + className + "->" + method + ":" + params
                + " " + methodReference.getReturnType());

        String record = className.substring(1, className.length() - 1).replace('/', '.') + ":" +
                method + ":" + params + ":" + methodReference.getReturnType();
        if (Build.usedMethods.contains(record)) {
            return;
        }
        if (p.matcher(className).find()) {
            return;
        }
        Build.usedMethods.add(record);
    }

    private static String getParamsType(List<? extends CharSequence> parameterTypes) {
        StringBuilder params = new StringBuilder();
        for (CharSequence charSequence : parameterTypes) {
            boolean isArray = false;
            String s = null;
            if (charSequence.toString().startsWith("[")) {
                s = charSequence.subSequence(1, charSequence.length()).toString();
                isArray = true;
            } else {
                s = charSequence.toString();
            }
            if (!APatchTool.mappingMap.containsValue(s)) {
                params.append(isArray ? "[" + s + "|" : s + "|");
                continue;
            } else {
                for (Map.Entry<String, String> entry : APatchTool.mappingMap.entrySet()) {
                    if (entry.getValue().equals(charSequence.toString())) {
                        params.append(isArray ? "[" + entry.getKey() + "|" : entry.getKey() + "|");
                    }
                }

            }
        }
        if (params.length() > 1) {
            return "[" + params.substring(0, params.length() - 1).toString() + "]";
        }else {
            return "[" + params.toString() +"]";
        }
    }

    public static void addUsedClass(String className) {
//        if (Build.addClasses.contains(className.substring(1, className.length() - 1).replace('/', '.'))){
//            return;
//        }
        if (className == null || className.length() <= 3) {
            return;
        }
        String clazz = className;
        if (clazz.charAt(0) == '[') {
            clazz = clazz.substring(1);
        }
        if (p.matcher(className).find()) {
            return;
        }

        Build.usedClasses.add(clazz.substring(1, clazz.length() - 1).replace('/', '.'));
    }


    /**
     * 将变动的信息写入到一个文件
     *
     * @param append
     * @param outFile
     */
    public void save(BundleDiffResult bundleDiffResult) throws IOException {
            List<BundleDiffResult.ClassDiff>classDiffs = new ArrayList<>();
            for (ClassDiffInfo classDiffInfo:classDiffInfoMap.values()){
                BundleDiffResult.ClassDiff classDiff = new BundleDiffResult.ClassDiff();
                classDiff.setClassName(classDiffInfo.getName());
                classDiff.setDiffType(classDiffInfo.getType());
                classDiffs.add(classDiff);
                if (classDiffInfo.getModifyFields().size() > 0){
                    List<BundleDiffResult.FieldDiff>fieldDiffs = new ArrayList<>();
                    for (FieldDiffInfo fieldDiffInfo:classDiffInfo.getModifyFields()) {
                        BundleDiffResult.FieldDiff fieldDiff = new BundleDiffResult.FieldDiff();
                        fieldDiffs.add(fieldDiff);
                        fieldDiff.setDiffType(fieldDiffInfo.getType());
                        fieldDiff.setFieldDesc(ReferenceUtil.getFieldDescriptor(fieldDiffInfo.getBackedField()));
                    }
                    classDiff.setFieldDiffs(fieldDiffs);
                }
                if (classDiffInfo.getModifyMethods().size() > 0){
                    List<BundleDiffResult.MethodDiff>methodDiffs = new ArrayList<>();
                    for (MethodDiffInfo methodDiffInfo:classDiffInfo.getModifyMethods()) {
                        BundleDiffResult.MethodDiff methodDiff = new BundleDiffResult.MethodDiff();
                        methodDiffs.add(methodDiff);
                        methodDiff.setDiffType(methodDiffInfo.getType());
                        methodDiff.setMethodDesc(ReferenceUtil.getMethodDescriptor(methodDiffInfo.getBackedMethod()));
                    }
                    classDiff.setMethodDiffs(methodDiffs);
                }
            }

        bundleDiffResult.setClassDiffs(classDiffs);
    }

    public void writeToFile(String bundleName, File diffFile, File diffJsonFile) throws IOException {
        JSONArray jsonArray = new JSONArray();
        if (diffJsonFile.exists()) {
            String content = FileUtils.readFileToString(diffJsonFile);
            jsonArray = JSONArray.parseArray(content);
        }
        JSONObject jsonObject = new JSONObject();
        jsonObject.put("bundleName", bundleName);
        if (addedMethods.size() > 0) {
            List<String> addMethods = new ArrayList<String>();
            List<MethodDiffInfoObject> diffInfoObjects = new ArrayList<MethodDiffInfoObject>();
            for (DexBackedMethod method : addedMethods) {
                addMethods.add("[add new Method:]" + method.getReturnType()
                        + "  " + method.getName() + "("
                        + Formater.formatStringList(method.getParameterTypes())
                        + ")  in Class:" + method.getDefiningClass());
                MethodDiffInfoObject diffInfoObject = new MethodDiffInfoObject();
                diffInfoObject.setDiffType(DiffType.ADD);
                diffInfoObject.setReturnType(method.getReturnType());
                diffInfoObject.setMethodDeclaration(method.getName() + "("
                        + Formater.formatStringList(method.getParameterTypes())
                        + ")");
                diffInfoObject.setInClass(method.getDefiningClass());
                diffInfoObjects.add(diffInfoObject);
            }
            FileUtils.writeLines(diffFile, addMethods, true);
            jsonObject.put("addedMethods", diffInfoObjects);
        }
        if (modifiedMethods.size() > 0) {
            List<String> modifyMethods = new ArrayList<String>();
            List<MethodDiffInfoObject> diffInfoObjects = new ArrayList<MethodDiffInfoObject>();

            for (DexBackedMethod method : modifiedMethods) {
                modifyMethods.add("[modify Method:]" + method.getReturnType()
                        + "  " + method.getName() + "("
                        + Formater.formatStringList(method.getParameterTypes())
                        + ")  in Class:" + method.getDefiningClass());
                MethodDiffInfoObject diffInfoObject = new MethodDiffInfoObject();
                diffInfoObject.setDiffType(DiffType.MODIFY);
                diffInfoObject.setReturnType(method.getReturnType());
                diffInfoObject.setMethodDeclaration(method.getName() + "("
                        + Formater.formatStringList(method.getParameterTypes())
                        + ")");
                diffInfoObject.setInClass(method.getDefiningClass());
                diffInfoObjects.add(diffInfoObject);
            }
            FileUtils.writeLines(diffFile, modifyMethods, true);
            jsonObject.put("modifiedMethods", diffInfoObjects);
        }
        jsonArray.add(jsonObject);
        FileUtils.writeStringToFile(diffJsonFile, JSON.toJSONString(jsonArray));
    }

    public Set<DexBackedClassDef> getOldClasses() {
        return oldClasses;
    }

    public Set<DexBackedClassDef> getNewClasses() {
        return newClasses;
    }

    public void setOldClasses(Set<? extends DexBackedClassDef> oldClasses) {
        this.oldClasses.addAll(oldClasses);
    }

    public void setNewClasses(Set<? extends DexBackedClassDef> newClasses) {
        this.newClasses.addAll(newClasses);
    }

    public static void release() {
        modifiedClasses.clear();
        addedMethods.clear();
        modifiedMethods.clear();
        addedClasses.clear();
    }


    public void update() throws PatchException {
        if (classDiffInfoMap.size() > 0) {
            for (ClassDiffInfo classDiffInfo : classDiffInfoMap.values()) {
                for (MethodDiffInfo methodDiffInfo : classDiffInfo.getModifyMethods()) {
                    if (methodDiffInfo.getType().equals(DiffType.MODIFY)) {
                        addModifiedMethods(methodDiffInfo.getBackedMethod());
                    }
                }
                if (classDiffInfo.getType().equals(DiffType.ADD)) {
                    setAddedClasses(classDiffInfo.getClassDef());
                    addManifestAddClass(classDiffInfo.getClassDef().getType());
                }
            }
        }
    }
}
