/*
 *
 *
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "[]"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2016 Alibaba Group
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 *
 *
 */


package com.taobao.common.dexpatcher;

import com.taobao.common.dexpatcher.algorithms.patch.*;
import com.taobao.common.dexpatcher.struct.DexPatchFile;
import com.taobao.common.dexpatcher.struct.SmallPatchedDexItemFile;
import com.taobao.dex.*;
import com.taobao.dex.util.CompareUtils;
import com.taobao.dx.util.Hex;
import com.taobao.dx.util.IndexMap;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;

public class DexPatchApplier {
    private final Dex oldDex;
    private final Dex patchedDex;

    /** May be null if we need to generate small patch. **/
    private final DexPatchFile patchFile;

    private final SmallPatchedDexItemFile extraInfoFile;
    private final IndexMap oldToFullPatchedIndexMap;
    private final IndexMap patchedToSmallPatchedIndexMap;

    private final String oldDexSignStr;

    private DexSectionPatchAlgorithm<StringData> stringDataSectionPatchAlg;
    private DexSectionPatchAlgorithm<Integer> typeIdSectionPatchAlg;
    private DexSectionPatchAlgorithm<ProtoId> protoIdSectionPatchAlg;
    private DexSectionPatchAlgorithm<FieldId> fieldIdSectionPatchAlg;
    private DexSectionPatchAlgorithm<MethodId> methodIdSectionPatchAlg;
    private DexSectionPatchAlgorithm<ClassDef> classDefSectionPatchAlg;
    private DexSectionPatchAlgorithm<TypeList> typeListSectionPatchAlg;
    private DexSectionPatchAlgorithm<AnnotationSetRefList> annotationSetRefListSectionPatchAlg;
    private DexSectionPatchAlgorithm<AnnotationSet> annotationSetSectionPatchAlg;
    private DexSectionPatchAlgorithm<ClassData> classDataSectionPatchAlg;
    private DexSectionPatchAlgorithm<Code> codeSectionPatchAlg;
    private DexSectionPatchAlgorithm<DebugInfoItem> debugInfoSectionPatchAlg;
    private DexSectionPatchAlgorithm<Annotation> annotationSectionPatchAlg;
    private DexSectionPatchAlgorithm<EncodedValue> encodedArraySectionPatchAlg;
    private DexSectionPatchAlgorithm<AnnotationsDirectory> annotationsDirectorySectionPatchAlg;

    public DexPatchApplier(File oldDexIn, File patchFileIn) throws IOException {
        this(
                new Dex(oldDexIn),
                (patchFileIn != null ? new DexPatchFile(patchFileIn) : null),
                null
        );
    }

    public DexPatchApplier(InputStream oldDexIn, InputStream patchFileIn) throws IOException {
        this(
                new Dex(oldDexIn),
                (patchFileIn != null ? new DexPatchFile(patchFileIn) : null),
                null
        );
    }

    public DexPatchApplier(InputStream oldDexIn, int initDexSize, InputStream patchFileIn) throws IOException {
        this(
                new Dex(oldDexIn, initDexSize),
                (patchFileIn != null ? new DexPatchFile(patchFileIn) : null),
                null
        );
    }

    public DexPatchApplier(
            File oldDexIn,
            File patchFileIn,
            SmallPatchedDexItemFile extraInfoFile
    ) throws IOException {
        this(
                new Dex(oldDexIn),
                (patchFileIn != null ? new DexPatchFile(patchFileIn) : null),
                extraInfoFile
        );
    }

    public DexPatchApplier(
            InputStream oldDexIn,
            InputStream patchFileIn,
            SmallPatchedDexItemFile extraInfoFile
    ) throws IOException {
        this(
                new Dex(oldDexIn),
                (patchFileIn != null ? new DexPatchFile(patchFileIn) : null),
                extraInfoFile
        );
    }

    public DexPatchApplier(
            InputStream oldDexIn,
            int initDexSize,
            InputStream patchFileIn,
            SmallPatchedDexItemFile extraInfoFile
    ) throws IOException {
        this(
                new Dex(oldDexIn, initDexSize),
                (patchFileIn != null ? new DexPatchFile(patchFileIn) : null),
                extraInfoFile
        );
    }

    public DexPatchApplier(
            Dex oldDexIn,
            DexPatchFile patchFileIn,
            SmallPatchedDexItemFile extraAddedDexElementsFile
    ) {
        this.oldDex = oldDexIn;
        this.oldDexSignStr = Hex.toHexString(oldDexIn.computeSignature(false));
        this.patchFile = patchFileIn;
        if (extraAddedDexElementsFile == null) {
            this.patchedDex = new Dex(patchFileIn.getPatchedDexSize());
        } else {
            this.patchedDex = new Dex(
                    extraAddedDexElementsFile.getPatchedDexSizeByOldDexSign(this.oldDexSignStr)
            );
        }
        this.oldToFullPatchedIndexMap = new IndexMap();
        this.patchedToSmallPatchedIndexMap = (extraAddedDexElementsFile != null ? new IndexMap() : null);
        this.extraInfoFile = extraAddedDexElementsFile;

        if ((patchFileIn == null) && (extraAddedDexElementsFile == null
                || !extraAddedDexElementsFile.isAffectedOldDex(this.oldDexSignStr))) {
            throw new UnsupportedOperationException(
                    "patchFileIn is null, and extraAddedDexElementFile"
                            + "(smallPatchInfo) is null or does not mention "
                            + "oldDexIn."
            );
        }
    }

    public void executeAndSaveTo(OutputStream out) throws IOException {
        // Before executing, we should check if this patch can be applied to
        // old dex we passed in.
        byte[] oldDexSign = this.oldDex.computeSignature(false);
        if (oldDexSign == null) {
            throw new IOException("failed to compute old dex's signature.");
        }

        if (this.patchFile != null) {
            byte[] oldDexSignInPatchFile = this.patchFile.getOldDexSignature();
            if (CompareUtils.uArrCompare(oldDexSign, oldDexSignInPatchFile) != 0) {
                throw new IOException(
                        String.format(
                                "old dex signature mismatch! expected: %s, actual: %s",
                                Arrays.toString(oldDexSign),
                                Arrays.toString(oldDexSignInPatchFile)
                        )
                );
            }
        }

        String oldDexSignStr = Hex.toHexString(oldDexSign);

        // Firstly, set sections' offset after patched, sort according to their offset so that
        // the dex lib of aosp can calculate section size.
        TableOfContents patchedToc = this.patchedDex.getTableOfContents();

        patchedToc.header.off = 0;
        patchedToc.header.size = 1;
        patchedToc.mapList.size = 1;

        if (extraInfoFile == null || !extraInfoFile.isAffectedOldDex(this.oldDexSignStr)) {
            patchedToc.stringIds.off
                    = this.patchFile.getPatchedStringIdSectionOffset();
            patchedToc.typeIds.off
                    = this.patchFile.getPatchedTypeIdSectionOffset();
            patchedToc.typeLists.off
                    = this.patchFile.getPatchedTypeListSectionOffset();
            patchedToc.protoIds.off
                    = this.patchFile.getPatchedProtoIdSectionOffset();
            patchedToc.fieldIds.off
                    = this.patchFile.getPatchedFieldIdSectionOffset();
            patchedToc.methodIds.off
                    = this.patchFile.getPatchedMethodIdSectionOffset();
            patchedToc.classDefs.off
                    = this.patchFile.getPatchedClassDefSectionOffset();
            patchedToc.mapList.off
                    = this.patchFile.getPatchedMapListSectionOffset();
            patchedToc.stringDatas.off
                    = this.patchFile.getPatchedStringDataSectionOffset();
            patchedToc.annotations.off
                    = this.patchFile.getPatchedAnnotationSectionOffset();
            patchedToc.annotationSets.off
                    = this.patchFile.getPatchedAnnotationSetSectionOffset();
            patchedToc.annotationSetRefLists.off
                    = this.patchFile.getPatchedAnnotationSetRefListSectionOffset();
            patchedToc.annotationsDirectories.off
                    = this.patchFile.getPatchedAnnotationsDirectorySectionOffset();
            patchedToc.encodedArrays.off
                    = this.patchFile.getPatchedEncodedArraySectionOffset();
            patchedToc.debugInfos.off
                    = this.patchFile.getPatchedDebugInfoSectionOffset();
            patchedToc.codes.off
                    = this.patchFile.getPatchedCodeSectionOffset();
            patchedToc.classDatas.off
                    = this.patchFile.getPatchedClassDataSectionOffset();
            patchedToc.fileSize
                    = this.patchFile.getPatchedDexSize();
        } else {
            patchedToc.stringIds.off
                    = this.extraInfoFile.getPatchedStringIdOffsetByOldDexSign(oldDexSignStr);
            patchedToc.typeIds.off
                    = this.extraInfoFile.getPatchedTypeIdOffsetByOldDexSign(oldDexSignStr);
            patchedToc.typeLists.off
                    = this.extraInfoFile.getPatchedTypeListOffsetByOldDexSign(oldDexSignStr);
            patchedToc.protoIds.off
                    = this.extraInfoFile.getPatchedProtoIdOffsetByOldDexSign(oldDexSignStr);
            patchedToc.fieldIds.off
                    = this.extraInfoFile.getPatchedFieldIdOffsetByOldDexSign(oldDexSignStr);
            patchedToc.methodIds.off
                    = this.extraInfoFile.getPatchedMethodIdOffsetByOldDexSign(oldDexSignStr);
            patchedToc.classDefs.off
                    = this.extraInfoFile.getPatchedClassDefOffsetByOldDexSign(oldDexSignStr);
            patchedToc.mapList.off
                    = this.extraInfoFile.getPatchedMapListOffsetByOldDexSign(oldDexSignStr);
            patchedToc.stringDatas.off
                    = this.extraInfoFile.getPatchedStringDataOffsetByOldDexSign(oldDexSignStr);
            patchedToc.annotations.off
                    = this.extraInfoFile.getPatchedAnnotationOffsetByOldDexSign(oldDexSignStr);
            patchedToc.annotationSets.off
                    = this.extraInfoFile.getPatchedAnnotationSetOffsetByOldDexSign(oldDexSignStr);
            patchedToc.annotationSetRefLists.off
                    = this.extraInfoFile.getPatchedAnnotationSetRefListOffsetByOldDexSign(oldDexSignStr);
            patchedToc.annotationsDirectories.off
                    = this.extraInfoFile.getPatchedAnnotationsDirectoryOffsetByOldDexSign(oldDexSignStr);
            patchedToc.encodedArrays.off
                    = this.extraInfoFile.getPatchedEncodedArrayOffsetByOldDexSign(oldDexSignStr);
            patchedToc.debugInfos.off
                    = this.extraInfoFile.getPatchedDebugInfoOffsetByOldDexSign(oldDexSignStr);
            patchedToc.codes.off
                    = this.extraInfoFile.getPatchedCodeOffsetByOldDexSign(oldDexSignStr);
            patchedToc.classDatas.off
                    = this.extraInfoFile.getPatchedClassDataOffsetByOldDexSign(oldDexSignStr);
            patchedToc.fileSize
                    = this.extraInfoFile.getPatchedDexSizeByOldDexSign(oldDexSignStr);
        }

        Arrays.sort(patchedToc.sections);

        patchedToc.computeSizesFromOffsets();

        // Secondly, run patch algorithms according to sections' dependencies.
        this.stringDataSectionPatchAlg = new StringDataSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.typeIdSectionPatchAlg = new TypeIdSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.protoIdSectionPatchAlg = new ProtoIdSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.fieldIdSectionPatchAlg = new FieldIdSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.methodIdSectionPatchAlg = new MethodIdSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.classDefSectionPatchAlg = new ClassDefSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.typeListSectionPatchAlg = new TypeListSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.annotationSetRefListSectionPatchAlg = new AnnotationSetRefListSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.annotationSetSectionPatchAlg = new AnnotationSetSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.classDataSectionPatchAlg = new ClassDataSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.codeSectionPatchAlg = new CodeSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.debugInfoSectionPatchAlg = new DebugInfoItemSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.annotationSectionPatchAlg = new AnnotationSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.encodedArraySectionPatchAlg = new StaticValueSectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );
        this.annotationsDirectorySectionPatchAlg = new AnnotationsDirectorySectionPatchAlgorithm(
                patchFile, oldDex, patchedDex, oldToFullPatchedIndexMap,
                patchedToSmallPatchedIndexMap, extraInfoFile
        );

        this.stringDataSectionPatchAlg.execute();
        this.typeIdSectionPatchAlg.execute();
        this.typeListSectionPatchAlg.execute();
        this.protoIdSectionPatchAlg.execute();
        this.fieldIdSectionPatchAlg.execute();
        this.methodIdSectionPatchAlg.execute();
        Runtime.getRuntime().gc();
        this.annotationSectionPatchAlg.execute();
        this.annotationSetSectionPatchAlg.execute();
        this.annotationSetRefListSectionPatchAlg.execute();
        this.annotationsDirectorySectionPatchAlg.execute();
        Runtime.getRuntime().gc();
        this.debugInfoSectionPatchAlg.execute();
        this.codeSectionPatchAlg.execute();
        Runtime.getRuntime().gc();
        this.classDataSectionPatchAlg.execute();
        this.encodedArraySectionPatchAlg.execute();
        this.classDefSectionPatchAlg.execute();
        Runtime.getRuntime().gc();

        // Thirdly, write header, mapList. Calculate and write patched dex's sign and checksum.
        Dex.Section headerOut = this.patchedDex.openSection(patchedToc.header.off);
        patchedToc.writeHeader(headerOut);

        Dex.Section mapListOut = this.patchedDex.openSection(patchedToc.mapList.off);
        patchedToc.writeMap(mapListOut);

        this.patchedDex.writeHashes();

        // Finally, write patched dex to file.
        this.patchedDex.writeTo(out);
    }

    public void executeAndSaveTo(File file) throws IOException {
        OutputStream os = null;
        try {
            os = new BufferedOutputStream(new FileOutputStream(file));
            executeAndSaveTo(os);
        } finally {
            if (os != null) {
                try {
                    os.close();
                } catch (Exception e) {
                    // ignored.
                }
            }
        }
    }
}
