/*
 *
 *
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "[]"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2016 Alibaba Group
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 *
 *
 */


package com.taobao.common.dexpatcher;

import com.taobao.common.dexpatcher.algorithms.diff.*;
import com.taobao.common.dexpatcher.struct.DexPatchFile;
import com.taobao.common.dexpatcher.struct.PatchOperation;
import com.taobao.dex.*;
import com.taobao.dex.io.DexDataBuffer;
import com.taobao.dx.util.IndexMap;
import org.apache.commons.io.FileUtils;

import java.io.*;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;



public class DexPatchGenerator {
    private static final String TAG = "DexPatchGenerator";

    private final Dex oldDex;
    private final Dex newDex;
    private final DexPatcherLogger logger = new DexPatcherLogger();
    private DexSectionDiffAlgorithm<StringData> stringDataSectionDiffAlg;
    private DexSectionDiffAlgorithm<Integer> typeIdSectionDiffAlg;
    private DexSectionDiffAlgorithm<ProtoId> protoIdSectionDiffAlg;
    private DexSectionDiffAlgorithm<FieldId> fieldIdSectionDiffAlg;
    private DexSectionDiffAlgorithm<MethodId> methodIdSectionDiffAlg;
    private DexSectionDiffAlgorithm<ClassDef> classDefSectionDiffAlg;
    private DexSectionDiffAlgorithm<TypeList> typeListSectionDiffAlg;
    private DexSectionDiffAlgorithm<AnnotationSetRefList> annotationSetRefListSectionDiffAlg;
    private DexSectionDiffAlgorithm<AnnotationSet> annotationSetSectionDiffAlg;
    private DexSectionDiffAlgorithm<ClassData> classDataSectionDiffAlg;
    private DexSectionDiffAlgorithm<Code> codeSectionDiffAlg;
    private DexSectionDiffAlgorithm<DebugInfoItem> debugInfoSectionDiffAlg;
    private DexSectionDiffAlgorithm<Annotation> annotationSectionDiffAlg;
    private DexSectionDiffAlgorithm<EncodedValue> encodedArraySectionDiffAlg;
    private DexSectionDiffAlgorithm<AnnotationsDirectory> annotationsDirectorySectionDiffAlg;
    private Set<String> additionalRemovingClassPatternSet;
    private int patchedHeaderOffset = 0;
    private int patchedStringIdsOffset = 0;
    private int patchedTypeIdsOffset = 0;
    private int patchedProtoIdsOffset = 0;
    private int patchedFieldIdsOffset = 0;
    private int patchedMethodIdsOffset = 0;
    private int patchedClassDefsOffset = 0;
    private int patchedTypeListsOffset = 0;
    private int patchedAnnotationItemsOffset = 0;
    private int patchedAnnotationSetItemsOffset = 0;
    private int patchedAnnotationSetRefListItemsOffset = 0;
    private int patchedAnnotationsDirectoryItemsOffset = 0;
    private int patchedDebugInfoItemsOffset = 0;
    private int patchedCodeItemsOffset = 0;
    private int patchedClassDataItemsOffset = 0;
    private int patchedStringDataItemsOffset = 0;
    private int patchedEncodedArrayItemsOffset = 0;
    private int patchedMapListOffset = 0;
    private int patchedDexSize = 0;

    public DexPatchGenerator(File oldDexFile, File newDexFile) throws IOException {
        this(new Dex(oldDexFile), new Dex(newDexFile));
    }

    /**
     * Notice: you should close inputstream manually.
     */
    public DexPatchGenerator(File oldDexFile, InputStream newDexStream) throws IOException {
        this(new Dex(oldDexFile), new Dex(newDexStream));
    }

    /**
     * Notice: you should close inputstream manually.
     */
    public DexPatchGenerator(InputStream oldDexStream, InputStream newDexStream) throws IOException {
        this(new Dex(oldDexStream), new Dex(newDexStream));
    }

    public DexPatchGenerator(Dex oldDex, Dex newDex) {
        this.oldDex = oldDex;
        this.newDex = newDex;

        IndexMap oldToNewIndexMap = new IndexMap();
        IndexMap oldToPatchedIndexMap = new IndexMap();
        IndexMap newToPatchedIndexMap = new IndexMap();
        IndexMap selfIndexMapForSkip = new IndexMap();

        additionalRemovingClassPatternSet = new HashSet<String>();

        this.stringDataSectionDiffAlg = new StringDataSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.typeIdSectionDiffAlg = new TypeIdSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.protoIdSectionDiffAlg = new ProtoIdSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.fieldIdSectionDiffAlg = new FieldIdSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.methodIdSectionDiffAlg = new MethodIdSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.classDefSectionDiffAlg = new ClassDefSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.typeListSectionDiffAlg = new TypeListSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.annotationSetRefListSectionDiffAlg = new AnnotationSetRefListSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.annotationSetSectionDiffAlg = new AnnotationSetSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.classDataSectionDiffAlg = new ClassDataSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.codeSectionDiffAlg = new CodeSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.debugInfoSectionDiffAlg = new DebugInfoItemSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.annotationSectionDiffAlg = new AnnotationSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.encodedArraySectionDiffAlg = new StaticValueSectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
        this.annotationsDirectorySectionDiffAlg = new AnnotationsDirectorySectionDiffAlgorithm(
                oldDex, newDex,
                oldToNewIndexMap,
                oldToPatchedIndexMap,
                newToPatchedIndexMap,
                selfIndexMapForSkip
        );
    }


    public void setLogger(DexPatcherLogger.IDexPatcherLogger logger) {
        this.logger.setLoggerImpl(logger);
    }

    public void executeAndSaveTo(File file) throws IOException {
        OutputStream os = null;
        try {
            os = new BufferedOutputStream(new FileOutputStream(file));
            executeAndSaveTo(os);
        } finally {
            if (os != null) {
                try {
                    os.close();
                } catch (Exception e) {
                    // ignored.
                }
            }
        }
        if (file.exists() && file.length() == 0L){
            FileUtils.deleteQuietly(file);
        }
    }

    public void executeAndSaveTo(OutputStream out) throws IOException {
        // Firstly, collect information of items we want to remove additionally
        // in new dex and set them to corresponding diff algorithm implementations.
        Pattern[] classNamePatterns = new Pattern[this.additionalRemovingClassPatternSet.size()];
        int classNamePatternCount = 0;
        for (String regExStr : this.additionalRemovingClassPatternSet) {
            classNamePatterns[classNamePatternCount++] = Pattern.compile(regExStr);
        }

        List<Integer> typeIdOfClassDefsToRemove = new ArrayList<Integer>(classNamePatternCount);
        List<Integer> offsetOfClassDatasToRemove = new ArrayList<Integer>(classNamePatternCount);
        for (ClassDef classDef : this.newDex.classDefs()) {
            String typeName = this.newDex.typeNames().get(classDef.typeIndex);
            for (Pattern pattern : classNamePatterns) {
                if (pattern.matcher(typeName).matches()) {
                    typeIdOfClassDefsToRemove.add(classDef.typeIndex);
                    offsetOfClassDatasToRemove.add(classDef.classDataOffset);
                    break;
                }
            }
        }

        ((ClassDefSectionDiffAlgorithm) this.classDefSectionDiffAlg)
                .setTypeIdOfClassDefsToRemove(typeIdOfClassDefsToRemove);
        ((ClassDataSectionDiffAlgorithm) this.classDataSectionDiffAlg)
                .setOffsetOfClassDatasToRemove(offsetOfClassDatasToRemove);

        // Then, run diff algorithms according to sections' dependencies.

        // Use size calculated by algorithms above or from dex file definition to
        // calculate sections' offset and patched dex size.

        // Calculate header and id sections size, so that we can work out
        // the base offset of typeLists Section.
        int patchedheaderSize = SizeOf.HEADER_ITEM;
        int patchedStringIdsSize = newDex.getTableOfContents().stringIds.size * SizeOf.STRING_ID_ITEM;
        int patchedTypeIdsSize = newDex.getTableOfContents().typeIds.size * SizeOf.TYPE_ID_ITEM;

        // Although simulatePatchOperation can calculate this value, since protoIds section
        // depends on typeLists section, we can't run protoIds Section's simulatePatchOperation
        // method so far. Instead we calculate protoIds section's size using information in newDex
        // directly.
        int patchedProtoIdsSize = newDex.getTableOfContents().protoIds.size * SizeOf.PROTO_ID_ITEM;

        int patchedFieldIdsSize = newDex.getTableOfContents().fieldIds.size * SizeOf.MEMBER_ID_ITEM;
        int patchedMethodIdsSize = newDex.getTableOfContents().methodIds.size * SizeOf.MEMBER_ID_ITEM;
        int patchedClassDefsSize = newDex.getTableOfContents().classDefs.size * SizeOf.CLASS_DEF_ITEM;

        int patchedIdSectionSize =
                patchedStringIdsSize
                        + patchedTypeIdsSize
                        + patchedProtoIdsSize
                        + patchedFieldIdsSize
                        + patchedMethodIdsSize
                        + patchedClassDefsSize;

        this.patchedHeaderOffset = 0;

        // The diff works on each sections obey such procedure:
        //  1. Execute diff algorithms to calculate indices of items we need to add, del and replace.
        //  2. Execute patch algorithm simulation to calculate indices and offsets mappings that is
        //  necessary to next section's diff works.

        // Immediately do the patch simulation so that we can know:
        //  1. Indices and offsets mapping between old dex and patched dex.
        //  2. Indices and offsets mapping between new dex and patched dex.
        // These information will be used to do next diff works.
        this.patchedStringIdsOffset = patchedHeaderOffset + patchedheaderSize;
        if (this.oldDex.getTableOfContents().stringIds.isElementFourByteAligned) {
            this.patchedStringIdsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedStringIdsOffset);
        }
        this.stringDataSectionDiffAlg.execute();
        this.patchedStringDataItemsOffset = patchedheaderSize + patchedIdSectionSize;
        if (this.oldDex.getTableOfContents().stringDatas.isElementFourByteAligned) {
            this.patchedStringDataItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedStringDataItemsOffset);
        }
        this.stringDataSectionDiffAlg.simulatePatchOperation(this.patchedStringDataItemsOffset);

        this.typeIdSectionDiffAlg.execute();
        this.patchedTypeIdsOffset = this.patchedStringIdsOffset + patchedStringIdsSize;
        if (this.oldDex.getTableOfContents().typeIds.isElementFourByteAligned) {
            this.patchedTypeIdsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedTypeIdsOffset);
        }
        this.typeIdSectionDiffAlg.simulatePatchOperation(this.patchedTypeIdsOffset);

        this.typeListSectionDiffAlg.execute();
        this.patchedTypeListsOffset
                = patchedheaderSize
                + patchedIdSectionSize
                + this.stringDataSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().typeLists.isElementFourByteAligned) {
            this.patchedTypeListsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedTypeListsOffset);
        }
        this.typeListSectionDiffAlg.simulatePatchOperation(this.patchedTypeListsOffset);

        this.protoIdSectionDiffAlg.execute();
        this.patchedProtoIdsOffset = this.patchedTypeIdsOffset + patchedTypeIdsSize;
        if (this.oldDex.getTableOfContents().protoIds.isElementFourByteAligned) {
            this.patchedProtoIdsOffset = SizeOf.roundToTimesOfFour(this.patchedProtoIdsOffset);
        }
        this.protoIdSectionDiffAlg.simulatePatchOperation(this.patchedProtoIdsOffset);

        this.fieldIdSectionDiffAlg.execute();
        this.patchedFieldIdsOffset = this.patchedProtoIdsOffset + patchedProtoIdsSize;
        if (this.oldDex.getTableOfContents().fieldIds.isElementFourByteAligned) {
            this.patchedFieldIdsOffset = SizeOf.roundToTimesOfFour(this.patchedFieldIdsOffset);
        }
        this.fieldIdSectionDiffAlg.simulatePatchOperation(this.patchedFieldIdsOffset);

        this.methodIdSectionDiffAlg.execute();
        this.patchedMethodIdsOffset = this.patchedFieldIdsOffset + patchedFieldIdsSize;
        if (this.oldDex.getTableOfContents().methodIds.isElementFourByteAligned) {
            this.patchedMethodIdsOffset = SizeOf.roundToTimesOfFour(this.patchedMethodIdsOffset);
        }
        this.methodIdSectionDiffAlg.simulatePatchOperation(this.patchedMethodIdsOffset);

        this.annotationSectionDiffAlg.execute();
        this.patchedAnnotationItemsOffset
                = this.patchedTypeListsOffset
                + this.typeListSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().annotations.isElementFourByteAligned) {
            this.patchedAnnotationItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedAnnotationItemsOffset);
        }
        this.annotationSectionDiffAlg.simulatePatchOperation(this.patchedAnnotationItemsOffset);

        this.annotationSetSectionDiffAlg.execute();
        this.patchedAnnotationSetItemsOffset
                = this.patchedAnnotationItemsOffset
                + this.annotationSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().annotationSets.isElementFourByteAligned) {
            this.patchedAnnotationSetItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedAnnotationSetItemsOffset);
        }
        this.annotationSetSectionDiffAlg.simulatePatchOperation(
                this.patchedAnnotationSetItemsOffset
        );

        this.annotationSetRefListSectionDiffAlg.execute();
        this.patchedAnnotationSetRefListItemsOffset
                = this.patchedAnnotationSetItemsOffset
                + this.annotationSetSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().annotationSetRefLists.isElementFourByteAligned) {
            this.patchedAnnotationSetRefListItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedAnnotationSetRefListItemsOffset);
        }
        this.annotationSetRefListSectionDiffAlg.simulatePatchOperation(
                this.patchedAnnotationSetRefListItemsOffset
        );

        this.annotationsDirectorySectionDiffAlg.execute();
        this.patchedAnnotationsDirectoryItemsOffset
                = this.patchedAnnotationSetRefListItemsOffset
                + this.annotationSetRefListSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().annotationsDirectories.isElementFourByteAligned) {
            this.patchedAnnotationsDirectoryItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedAnnotationsDirectoryItemsOffset);
        }
        this.annotationsDirectorySectionDiffAlg.simulatePatchOperation(
                this.patchedAnnotationsDirectoryItemsOffset
        );

        this.debugInfoSectionDiffAlg.execute();
        this.patchedDebugInfoItemsOffset
                = this.patchedAnnotationsDirectoryItemsOffset
                + this.annotationsDirectorySectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().debugInfos.isElementFourByteAligned) {
            this.patchedDebugInfoItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedDebugInfoItemsOffset);
        }
        this.debugInfoSectionDiffAlg.simulatePatchOperation(this.patchedDebugInfoItemsOffset);

        this.codeSectionDiffAlg.execute();
        this.patchedCodeItemsOffset
                = this.patchedDebugInfoItemsOffset
                + this.debugInfoSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().codes.isElementFourByteAligned) {
            this.patchedCodeItemsOffset = SizeOf.roundToTimesOfFour(this.patchedCodeItemsOffset);
        }
        this.codeSectionDiffAlg.simulatePatchOperation(this.patchedCodeItemsOffset);

        this.classDataSectionDiffAlg.execute();
        this.patchedClassDataItemsOffset
                = this.patchedCodeItemsOffset
                + this.codeSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().classDatas.isElementFourByteAligned) {
            this.patchedClassDataItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedClassDataItemsOffset);
        }
        this.classDataSectionDiffAlg.simulatePatchOperation(this.patchedClassDataItemsOffset);

        this.encodedArraySectionDiffAlg.execute();
        this.patchedEncodedArrayItemsOffset
                = this.patchedClassDataItemsOffset
                + this.classDataSectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().encodedArrays.isElementFourByteAligned) {
            this.patchedEncodedArrayItemsOffset
                    = SizeOf.roundToTimesOfFour(this.patchedEncodedArrayItemsOffset);
        }
        this.encodedArraySectionDiffAlg.simulatePatchOperation(this.patchedEncodedArrayItemsOffset);

        this.classDefSectionDiffAlg.execute();
        this.patchedClassDefsOffset = this.patchedMethodIdsOffset + patchedMethodIdsSize;
        if (this.oldDex.getTableOfContents().classDefs.isElementFourByteAligned) {
            this.patchedClassDefsOffset = SizeOf.roundToTimesOfFour(this.patchedClassDefsOffset);
        }

        // Calculate any values we still know nothing about them.
        this.patchedMapListOffset
                = this.patchedEncodedArrayItemsOffset
                + this.encodedArraySectionDiffAlg.getPatchedSectionSize();
        if (this.oldDex.getTableOfContents().mapList.isElementFourByteAligned) {
            this.patchedMapListOffset = SizeOf.roundToTimesOfFour(this.patchedMapListOffset);
        }
        int patchedMapListSize = newDex.getTableOfContents().mapList.byteCount;

        this.patchedDexSize
                = this.patchedMapListOffset
                + patchedMapListSize;

        // Finally, write results to patch file.
        writeResultToStream(out);
    }

    private void writeResultToStream(OutputStream os) throws IOException {
        if (stringDataSectionDiffAlg.getPatchOperationList().size() == 0 &&
        typeIdSectionDiffAlg.getPatchOperationList().size() == 0 &&
                typeListSectionDiffAlg.getPatchOperationList().size() == 0&&
                protoIdSectionDiffAlg.getPatchOperationList().size() == 0 &&
        fieldIdSectionDiffAlg.getPatchOperationList().size() == 0&&
        methodIdSectionDiffAlg.getPatchOperationList().size() == 0&&
        annotationSectionDiffAlg.getPatchOperationList().size() == 0&&
        annotationSetSectionDiffAlg.getPatchOperationList().size() == 0&&
        annotationSetRefListSectionDiffAlg.getPatchOperationList().size() == 0&&
        annotationsDirectorySectionDiffAlg.getPatchOperationList().size() == 0&&
        debugInfoSectionDiffAlg.getPatchOperationList().size() == 0&&
        codeSectionDiffAlg.getPatchOperationList().size() == 0&&
        classDataSectionDiffAlg.getPatchOperationList().size() == 0&&
        encodedArraySectionDiffAlg.getPatchOperationList().size() == 0&&
        classDefSectionDiffAlg.getPatchOperationList().size() == 0){
            return;
        }
        DexDataBuffer buffer = new DexDataBuffer();
        buffer.write(DexPatchFile.MAGIC);
        buffer.writeShort(DexPatchFile.CURRENT_VERSION);
        buffer.writeInt(this.patchedDexSize);
        // we will return here to write firstChunkOffset later.
        int posOfFirstChunkOffsetField = buffer.position();
        buffer.writeInt(0);
        buffer.writeInt(this.patchedStringIdsOffset);
        buffer.writeInt(this.patchedTypeIdsOffset);
        buffer.writeInt(this.patchedProtoIdsOffset);
        buffer.writeInt(this.patchedFieldIdsOffset);
        buffer.writeInt(this.patchedMethodIdsOffset);
        buffer.writeInt(this.patchedClassDefsOffset);
        buffer.writeInt(this.patchedMapListOffset);
        buffer.writeInt(this.patchedTypeListsOffset);
        buffer.writeInt(this.patchedAnnotationSetRefListItemsOffset);
        buffer.writeInt(this.patchedAnnotationSetItemsOffset);
        buffer.writeInt(this.patchedClassDataItemsOffset);
        buffer.writeInt(this.patchedCodeItemsOffset);
        buffer.writeInt(this.patchedStringDataItemsOffset);
        buffer.writeInt(this.patchedDebugInfoItemsOffset);
        buffer.writeInt(this.patchedAnnotationItemsOffset);
        buffer.writeInt(this.patchedEncodedArrayItemsOffset);
        buffer.writeInt(this.patchedAnnotationsDirectoryItemsOffset);
        buffer.write(this.oldDex.computeSignature(false));
        int firstChunkOffset = buffer.position();
        buffer.position(posOfFirstChunkOffsetField);
        buffer.writeInt(firstChunkOffset);
        buffer.position(firstChunkOffset);

        writePatchOperations(buffer, this.stringDataSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.typeIdSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.typeListSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.protoIdSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.fieldIdSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.methodIdSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.annotationSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.annotationSetSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.annotationSetRefListSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.annotationsDirectorySectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.debugInfoSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.codeSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.classDataSectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.encodedArraySectionDiffAlg.getPatchOperationList());
        writePatchOperations(buffer, this.classDefSectionDiffAlg.getPatchOperationList());

        byte[] bufferData = buffer.array();
        os.write(bufferData);
        os.flush();
    }

    private <T extends Comparable<T>> void writePatchOperations(
            DexDataBuffer buffer, List<PatchOperation<T>> patchOperationList
    ) {
        List<Integer> delOpIndexList = new ArrayList<Integer>(patchOperationList.size());
        List<Integer> addOpIndexList = new ArrayList<Integer>(patchOperationList.size());
        List<Integer> replaceOpIndexList = new ArrayList<Integer>(patchOperationList.size());
        List<T> newItemList = new ArrayList<T>(patchOperationList.size());

        for (PatchOperation<T> patchOperation : patchOperationList) {
            switch (patchOperation.op) {
                case PatchOperation.OP_DEL: {
                    delOpIndexList.add(patchOperation.index);
                    break;
                }
                case PatchOperation.OP_ADD: {
                    addOpIndexList.add(patchOperation.index);
                    newItemList.add(patchOperation.newItem);
                    break;
                }
                case PatchOperation.OP_REPLACE: {
                    replaceOpIndexList.add(patchOperation.index);
                    newItemList.add(patchOperation.newItem);
                    break;
                }
            }
        }

        buffer.writeUleb128(delOpIndexList.size());
        int lastIndex = 0;
        for (Integer index : delOpIndexList) {
            buffer.writeSleb128(index - lastIndex);
            lastIndex = index;
        }

        buffer.writeUleb128(addOpIndexList.size());
        lastIndex = 0;
        for (Integer index : addOpIndexList) {
            buffer.writeSleb128(index - lastIndex);
            lastIndex = index;
        }

        buffer.writeUleb128(replaceOpIndexList.size());
        lastIndex = 0;
        for (Integer index : replaceOpIndexList) {
            buffer.writeSleb128(index - lastIndex);
            lastIndex = index;
        }

        for (T newItem : newItemList) {
            if (newItem instanceof StringData) {
                buffer.writeStringData((StringData) newItem);
            } else
            if (newItem instanceof Integer) {
                // TypeId item.
                buffer.writeInt((Integer) newItem);
            } else
            if (newItem instanceof TypeList) {
                buffer.writeTypeList((TypeList) newItem);
            } else
            if (newItem instanceof ProtoId) {
                buffer.writeProtoId((ProtoId) newItem);
            } else
            if (newItem instanceof FieldId) {
                buffer.writeFieldId((FieldId) newItem);
            } else
            if (newItem instanceof MethodId) {
                buffer.writeMethodId((MethodId) newItem);
            } else
            if (newItem instanceof Annotation) {
                buffer.writeAnnotation((Annotation) newItem);
            } else
            if (newItem instanceof AnnotationSet) {
                buffer.writeAnnotationSet((AnnotationSet) newItem);
            } else
            if (newItem instanceof AnnotationSetRefList) {
                buffer.writeAnnotationSetRefList(
                        (AnnotationSetRefList) newItem
                );
            } else
            if (newItem instanceof AnnotationsDirectory) {
                buffer.writeAnnotationsDirectory(
                        (AnnotationsDirectory) newItem
                );
            } else
            if (newItem instanceof DebugInfoItem) {
                buffer.writeDebugInfoItem((DebugInfoItem) newItem);
            } else
            if (newItem instanceof Code) {
                buffer.writeCode((Code) newItem);
            } else
            if (newItem instanceof ClassData) {
                buffer.writeClassData((ClassData) newItem);
            } else
            if (newItem instanceof EncodedValue) {
                buffer.writeEncodedArray((EncodedValue) newItem);
            } else
            if (newItem instanceof ClassDef) {
                buffer.writeClassDef((ClassDef) newItem);
            } else {
                throw new IllegalStateException(
                        "Unknown item type: " + newItem.getClass()
                );
            }
        }
    }
}
