/*
 *
 *
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "[]"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2016 Alibaba Group
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 *
 *
 */


package com.taobao.common.dexpatcher.algorithms.patch;

import com.taobao.common.dexpatcher.struct.DexPatchFile;
import com.taobao.dex.Dex;
import com.taobao.dex.SizeOf;
import com.taobao.dex.TableOfContents;
import com.taobao.dex.io.DexDataBuffer;
import com.taobao.dx.util.Hex;
import com.taobao.dx.util.IndexMap;

import java.util.Arrays;


public abstract class DexSectionPatchAlgorithm<T extends Comparable<T>> {
    protected final DexPatchFile patchFile;

    protected final Dex oldDex;

    /**
     * IndexMap for mapping old item to corresponding one in full patch.
     */
    private final IndexMap oldToFullPatchedIndexMap;

    /**
     * IndexMap for mapping item in full patch to corresponding one in small patch.
     */
    private final IndexMap fullPatchedToSmallPatchedIndexMap;

    /**
     * Signature string of dex we're processing. For extra info file usage.
     */
    private final String oldDexSignStr;
    private SmallPatchedDexItemChooser smallPatchedDexItemChooser = null;

    public DexSectionPatchAlgorithm(
            DexPatchFile patchFile,
            Dex oldDex,
            IndexMap oldToFullPatchedIndexMap,
            IndexMap fullPatchedToSmallPatchedIndexMap
    ) {
        this(patchFile, oldDex, oldToFullPatchedIndexMap, fullPatchedToSmallPatchedIndexMap, null);
    }

    public DexSectionPatchAlgorithm(
            DexPatchFile patchFile,
            Dex oldDex,
            IndexMap oldToFullPatchedIndexMap,
            IndexMap fullPatchedToSmallPatchedIndexMap,
            SmallPatchedDexItemChooser smallPatchedDexItemChooser
    ) {
        this.patchFile = patchFile;
        this.oldDex = oldDex;
        this.oldToFullPatchedIndexMap = oldToFullPatchedIndexMap;
        this.fullPatchedToSmallPatchedIndexMap = fullPatchedToSmallPatchedIndexMap;
        this.oldDexSignStr = Hex.toHexString(oldDex.computeSignature(false));
        this.smallPatchedDexItemChooser = smallPatchedDexItemChooser;
    }

    /**
     * Get {@link TableOfContents.Section} from {@code dex}.
     */
    protected abstract TableOfContents.Section getTocSection(Dex dex);

    /**
     * Get next item in {@code section}.
     */
    protected abstract T nextItem(DexDataBuffer section);

    /**
     * Get size of {@code item}.
     */
    protected abstract int getItemSize(T item);

    /**
     * Adjust {@code item} using specific {@code indexMap}
     */
    protected T adjustItem(IndexMap indexMap, T item) {
        return item;
    }

    /**
     * Update index or offset mapping in {@code indexMap}.
     */
    protected void updateIndexOrOffset(IndexMap indexMap, int oldIndex, int oldOffset, int newIndex, int newOffset) {
        // Should override by subclass if needed.
    }

    /**
     * Mark deleted index or offset in {@code indexMap}.
     *
     * Here we mark deleted item for such a case like this:
     *   Item in DebugInfo section reference a string in StringData section
     *   by index X, while in patched dex, the referenced string is removed.
     *
     * The {@code indexMap} must be aware of this case and return -1
     * instead of the original value X.
     *
     * Further more, the special value -1 is not chosen by our inspiration but
     * the definition of NO_INDEX in document of dex file format.
     */
    protected void markDeletedIndexOrOffset(IndexMap indexMap, int deletedIndex, int deletedOffset) {
        // Should override by subclass if needed.
    }

    /**
     * Judge if item on index {@code patchedIndex} should be kept in small dex.
     */
    protected final boolean isPatchedItemInSmallPatchedDex(String oldDexSignStr, int patchedIndex) {
        if (this.smallPatchedDexItemChooser != null) {
            return this.smallPatchedDexItemChooser
                    .isPatchedItemInSmallPatchedDex(oldDexSignStr, patchedIndex);
        } else {
            return true;
        }
    }

    /**
     * Return base offset of current section in full patched dex.
     */
    protected abstract int getFullPatchSectionBase();

    /**
     * Output patched item. This method should be overrided by subclass
     * so that patched item can be written to right place.
     * <p/>
     * Returns the offset where {@code patchedItem} is written. (<b>Must be valid.</b>)
     */
    protected abstract int writePatchedItem(T patchedItem);

    private int[] readDeltaIndiciesOrOffsets(int count) {
        int[] result = new int[count];
        int lastVal = 0;
        for (int i = 0; i < count; ++i) {
            int delta = patchFile.getBuffer().readSleb128();
            lastVal = lastVal + delta;
            result[i] = lastVal;
        }
        return result;
    }

    /**
     * Adapter method for item's offset fetching, if an item is not
     * inherited from {@link TableOfContents.Section.Item} (which means it is a simple item in dex section
     * that doesn't need multiple members to describe), this method
     * return {@code index} instead.
     */
    private int getItemOffsetOrIndex(int index, T item) {
        if (item instanceof TableOfContents.Section.Item) {
            return ((TableOfContents.Section.Item<?>) item).off;
        } else {
            return index;
        }
    }

    public void execute() {
        int deletedItemCount;
        int[] deletedIndices;

        int addedItemCount;
        int[] addedIndices;

        int replacedItemCount;
        int[] replacedIndices;

        if (patchFile != null) {
            deletedItemCount = patchFile.getBuffer().readUleb128();
            deletedIndices = readDeltaIndiciesOrOffsets(deletedItemCount);

            addedItemCount = patchFile.getBuffer().readUleb128();
            addedIndices = readDeltaIndiciesOrOffsets(addedItemCount);

            replacedItemCount = patchFile.getBuffer().readUleb128();
            replacedIndices = readDeltaIndiciesOrOffsets(replacedItemCount);
        } else {
            deletedItemCount = 0;
            deletedIndices = new int[deletedItemCount];

            addedItemCount = 0;
            addedIndices = new int[addedItemCount];

            replacedItemCount = 0;
            replacedIndices = new int[replacedItemCount];
        }

        TableOfContents.Section tocSec = getTocSection(this.oldDex);
        Dex.Section oldSection = null;

        int oldItemCount = 0;
        if (tocSec.exists()) {
            oldSection = this.oldDex.openSection(tocSec);
            oldItemCount = tocSec.size;
        }

        // Now rest data are added and replaced items arranged in the order of
        // added indices and replaced indices.
        boolean genFullPatchDex = (fullPatchedToSmallPatchedIndexMap == null);

        if (genFullPatchDex) {
            doFullPatch(
                    oldSection, oldItemCount, deletedIndices, addedIndices, replacedIndices
            );
        } else {
            doSmallPatch(
                    oldSection, oldItemCount, deletedIndices, addedIndices, replacedIndices
            );
        }
    }

    private void doFullPatch(
            Dex.Section oldSection,
            int oldItemCount,
            int[] deletedIndices,
            int[] addedIndices,
            int[] replacedIndices
    ) {
        int deletedItemCount = deletedIndices.length;
        int addedItemCount = addedIndices.length;
        int replacedItemCount = replacedIndices.length;
        int newItemCount = oldItemCount + addedItemCount - deletedItemCount;

        int deletedItemCounter = 0;
        int addActionCursor = 0;
        int replaceActionCursor = 0;

        int oldIndex = 0;
        int patchedIndex = 0;
        while (oldIndex < oldItemCount || patchedIndex < newItemCount) {
            if (addActionCursor < addedItemCount && addedIndices[addActionCursor] == patchedIndex) {
                T addedItem = nextItem(patchFile.getBuffer());
                int patchedOffset = writePatchedItem(addedItem);
                ++addActionCursor;
                ++patchedIndex;
            } else
            if (replaceActionCursor < replacedItemCount && replacedIndices[replaceActionCursor] == patchedIndex) {
                T replacedItem = nextItem(patchFile.getBuffer());
                int patchedOffset = writePatchedItem(replacedItem);
                ++replaceActionCursor;
                ++patchedIndex;
            } else
            if (Arrays.binarySearch(deletedIndices, oldIndex) >= 0) {
                T skippedOldItem = nextItem(oldSection); // skip old item.
                markDeletedIndexOrOffset(
                        oldToFullPatchedIndexMap,
                        oldIndex,
                        getItemOffsetOrIndex(oldIndex, skippedOldItem)
                );
                ++oldIndex;
                ++deletedItemCounter;
            } else
            if (Arrays.binarySearch(replacedIndices, oldIndex) >= 0) {
                T skippedOldItem = nextItem(oldSection); // skip old item.
                markDeletedIndexOrOffset(
                        oldToFullPatchedIndexMap,
                        oldIndex,
                        getItemOffsetOrIndex(oldIndex, skippedOldItem)
                );
                ++oldIndex;
            } else
            if (oldIndex < oldItemCount) {
                T oldItem = adjustItem(this.oldToFullPatchedIndexMap, nextItem(oldSection));

                int patchedOffset = writePatchedItem(oldItem);

                updateIndexOrOffset(
                        this.oldToFullPatchedIndexMap,
                        oldIndex,
                        getItemOffsetOrIndex(oldIndex, oldItem),
                        patchedIndex,
                        patchedOffset
                );

                ++oldIndex;
                ++patchedIndex;
            }
        }

        if (addActionCursor != addedItemCount || deletedItemCounter != deletedItemCount
                || replaceActionCursor != replacedItemCount
        ) {
            throw new IllegalStateException(
                    String.format(
                            "bad patch operation sequence. addCounter: %d, addCount: %d, "
                                    + "delCounter: %d, delCount: %d, "
                                    + "replaceCounter: %d, replaceCount:%d",
                            addActionCursor,
                            addedItemCount,
                            deletedItemCounter,
                            deletedItemCount,
                            replaceActionCursor,
                            replacedItemCount
                    )
            );
        }
    }

    private void doSmallPatch(
            Dex.Section oldSection,
            int oldItemCount,
            int[] deletedIndices,
            int[] addedIndices,
            int[] replacedIndices
    ) {
        int deletedItemCount = deletedIndices.length;
        int addedItemCount = addedIndices.length;
        int replacedItemCount = replacedIndices.length;
        int newItemCount = oldItemCount + addedItemCount - deletedItemCount;

        int deletedItemCounter = 0;
        int addActionCursor = 0;
        int replaceActionCursor = 0;

        int oldIndex = 0;
        int fullPatchedIndex = 0;
        int fullPatchedOffset = getFullPatchSectionBase();
        int smallPatchedIndex = 0;
        while (oldIndex < oldItemCount || fullPatchedIndex < newItemCount) {
            if (addActionCursor < addedItemCount && addedIndices[addActionCursor] == fullPatchedIndex) {
                T addedItem = nextItem(patchFile.getBuffer());
                ++addActionCursor;

                if (getTocSection(oldDex).isElementFourByteAligned) {
                    fullPatchedOffset = SizeOf.roundToTimesOfFour(fullPatchedOffset);
                }

                if (isPatchedItemInSmallPatchedDex(this.oldDexSignStr, fullPatchedIndex)) {
                    T adjustedItem = adjustItem(fullPatchedToSmallPatchedIndexMap, addedItem);
                    int smallPatchedOffset = writePatchedItem(adjustedItem);
                    updateIndexOrOffset(
                            fullPatchedToSmallPatchedIndexMap,
                            fullPatchedIndex,
                            fullPatchedOffset,
                            smallPatchedIndex,
                            smallPatchedOffset
                    );
                    ++smallPatchedIndex;
                }

                ++fullPatchedIndex;
                fullPatchedOffset += getItemSize(addedItem);
            } else
            if (replaceActionCursor < replacedItemCount && replacedIndices[replaceActionCursor] == fullPatchedIndex) {
                T replacedItem = nextItem(patchFile.getBuffer());
                ++replaceActionCursor;

                if (getTocSection(oldDex).isElementFourByteAligned) {
                    fullPatchedOffset = SizeOf.roundToTimesOfFour(fullPatchedOffset);
                }

                if (isPatchedItemInSmallPatchedDex(this.oldDexSignStr, fullPatchedIndex)) {
                    T adjustedItem = adjustItem(fullPatchedToSmallPatchedIndexMap, replacedItem);
                    int smallPatchedOffset = writePatchedItem(adjustedItem);
                    updateIndexOrOffset(
                            fullPatchedToSmallPatchedIndexMap,
                            fullPatchedIndex,
                            fullPatchedOffset,
                            smallPatchedIndex,
                            smallPatchedOffset
                    );
                    ++smallPatchedIndex;
                }

                ++fullPatchedIndex;
                fullPatchedOffset += getItemSize(replacedItem);
            } else
            if (Arrays.binarySearch(deletedIndices, oldIndex) >= 0) {
                T skippedOldItem = nextItem(oldSection); // skip old item.
                markDeletedIndexOrOffset(
                        oldToFullPatchedIndexMap,
                        oldIndex,
                        getItemOffsetOrIndex(oldIndex, skippedOldItem)
                );
                ++oldIndex;
                ++deletedItemCounter;
            } else
            if (Arrays.binarySearch(replacedIndices, oldIndex) >= 0) {
                T skippedOldItem = nextItem(oldSection); // skip old item.
                markDeletedIndexOrOffset(
                        oldToFullPatchedIndexMap,
                        oldIndex,
                        getItemOffsetOrIndex(oldIndex, skippedOldItem)
                );
                ++oldIndex;
            } else
            if (oldIndex < oldItemCount) {
                T oldItem = nextItem(oldSection);
                T oldItemInFullPatch = adjustItem(this.oldToFullPatchedIndexMap, oldItem);

                if (getTocSection(oldDex).isElementFourByteAligned) {
                    fullPatchedOffset = SizeOf.roundToTimesOfFour(fullPatchedOffset);
                }

                if (isPatchedItemInSmallPatchedDex(this.oldDexSignStr, fullPatchedIndex)) {
                    T patchedItemInSmallPatch = adjustItem(
                            this.fullPatchedToSmallPatchedIndexMap, oldItemInFullPatch
                    );
                    int smallPatchedOffset = writePatchedItem(patchedItemInSmallPatch);
                    updateIndexOrOffset(
                            fullPatchedToSmallPatchedIndexMap,
                            fullPatchedIndex,
                            fullPatchedOffset,
                            smallPatchedIndex,
                            smallPatchedOffset
                    );
                    ++smallPatchedIndex;
                }

                updateIndexOrOffset(
                        oldToFullPatchedIndexMap,
                        oldIndex,
                        getItemOffsetOrIndex(oldIndex, oldItem),
                        fullPatchedIndex,
                        fullPatchedOffset
                );

                ++fullPatchedIndex;
                fullPatchedOffset += getItemSize(oldItemInFullPatch);

                ++oldIndex;
            }
        }

        if (addActionCursor != addedItemCount || deletedItemCounter != deletedItemCount
                || replaceActionCursor != replacedItemCount
        ) {
            throw new IllegalStateException(
                    String.format(
                            "bad patch operation sequence. addCounter: %d, addCount: %d, "
                                    + "delCounter: %d, delCount: %d, "
                                    + "replaceCounter: %d, replaceCount:%d",
                            addActionCursor,
                            addedItemCount,
                            deletedItemCounter,
                            deletedItemCount,
                            replaceActionCursor,
                            replacedItemCount
                    )
            );
        }
    }

    /**
     * Indicates if an item in full patched dex with specific index
     * should be kept in small patched dex of current old dex.
     */
    public interface SmallPatchedDexItemChooser {
        boolean isPatchedItemInSmallPatchedDex(String oldDexSign, int patchedItemIndex);
    }
}
